# -*- coding: utf-8 -*-
import shutil
import tempfile, os, base64
import unittest
from sqlalchemy.schema import Column
from sqlalchemy.types import Unicode, Integer
from .base_sqla import setup_database, clear_database, DeclarativeBase, DBSession
from .utils import OpenFiles
from depot.fields.sqlalchemy import UploadedFileField
from depot.fields.upload import UploadedFile
from depot.manager import DepotManager
from depot.fields.interfaces import FileFilter

from depot._compat import u_, bytes_

of = OpenFiles()

def setUpModule():
    setup_database()

    DepotManager._clear()
    DepotManager.configure('default', {'depot.storage_path': './lfs'})


def tearDownModule():
    shutil.rmtree('./lfs', ignore_errors=True)


class DictLikeCheckFilter(FileFilter):
    def on_save(self, uploaded_file):
        uploaded_file['changed'] = 'OLD'
        uploaded_file['hello'] = 'World'
        uploaded_file['deleted'] = False
        del uploaded_file['deleted']
        uploaded_file['changed'] = 'NEW'
        uploaded_file['deleted_attr'] = False
        del uploaded_file.deleted_attr

        try:
            del uploaded_file.missing_attr
        except AttributeError:
            uploaded_file.missing_attr_trapped = True


class SimpleDocument(DeclarativeBase):
    __tablename__ = 'simple_docu'

    uid = Column(Integer, autoincrement=True, primary_key=True)
    name = Column(Unicode(16), unique=True)
    content = Column('content_col', UploadedFileField([DictLikeCheckFilter()]))


class TestFieldsInterface(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.file_content = b'this is the file content'
        cls.fake_file = tempfile.NamedTemporaryFile()
        cls.fake_file.write(cls.file_content)
        cls.fake_file.flush()

    @classmethod
    def tearDownClass(cls):
        cls.fake_file.close()

    def setUp(self):
        clear_database()

    def tearDown(self):
        of.close_all()

    def test_column_is_dictlike(self):
        doc = SimpleDocument(name=u_('Foo'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        DBSession.flush()
        DBSession.commit()

        d = DBSession.query(SimpleDocument).filter_by(name=u_('Foo')).first()
        assert doc.content.hello == 'World', doc.content
        assert 'deleted' not in doc.content, doc.content
        assert doc.content.changed == 'NEW', doc.content
        assert 'deleted_attr' not in doc.content, doc.content
        assert doc.content.missing_attr_trapped == True, doc.content

    def test_column_cannot_edit_after_save(self):
        doc = SimpleDocument(name=u_('Foo'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        DBSession.flush()
        DBSession.commit()

        d = DBSession.query(SimpleDocument).filter_by(name=u_('Foo')).first()
        with self.assertRaises(TypeError):
            d.content['hello'] = 'Everybody'

    def test_column_cannot_delete_after_save(self):
        doc = SimpleDocument(name=u_('Foo'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        DBSession.flush()
        DBSession.commit()

        d = DBSession.query(SimpleDocument).filter_by(name=u_('Foo')).first()
        with self.assertRaises(TypeError):
            del d.content['hello']

    def test_column_cannot_edit_attr_after_save(self):
        doc = SimpleDocument(name=u_('Foo'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        DBSession.flush()
        DBSession.commit()

        d = DBSession.query(SimpleDocument).filter_by(name=u_('Foo')).first()
        with self.assertRaises(TypeError):
            d.content.hello = 'Everybody'

    def test_column_cannot_delete_attr_after_save(self):
        doc = SimpleDocument(name=u_('Foo'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        DBSession.flush()
        DBSession.commit()

        d = DBSession.query(SimpleDocument).filter_by(name=u_('Foo')).first()
        with self.assertRaises(TypeError):
            del d.content.hello

    def test_storage_does_not_exists(self):
        doc = SimpleDocument(name=u_('Foo'))
        with self.assertRaises(ValueError):
            doc.content = UploadedFile(of.open(self.fake_file.name, 'rb'),
                                       'missing_storage')
        DBSession.add(doc)
        DBSession.flush()
        DBSession.commit()



"""
Provides FileStorage implementation for Amazon S3.

This is useful for storing files in S3.

"""
from __future__ import absolute_import

from datetime import datetime
import warnings
import uuid
from configparser import DuplicateSectionError
import boto
from boto.s3.connection import S3Connection
from depot._compat import unicode_text, percent_encode, percent_decode
from depot.utils import make_content_disposition

from .interfaces import FileStorage, StoredFile
from . import utils

CANNED_ACL_PUBLIC_READ = 'public-read'
CANNED_ACL_PRIVATE = 'private'


class S3StoredFile(StoredFile):
    def __init__(self, file_id, key):
        _check_file_id(file_id)
        self._key = key
        filename = key.metadata.get('x-depot-filename')
        if filename:
            filename = percent_decode(filename)
        metadata_info = {'filename': filename,
                         'content_type': key.content_type,
                         'content_length': key.size,
                         'last_modified': None}

        try:
            last_modified = key.get_metadata('x-depot-modified')
            if last_modified:
                metadata_info['last_modified'] = datetime.strptime(last_modified,
                                                                   '%Y-%m-%d %H:%M:%S')
        except:
            pass

        super(S3StoredFile, self).__init__(file_id=file_id, **metadata_info)

    def read(self, n=-1):
        if self.closed:
            raise ValueError("cannot read from a closed file")

        if n <= 0:
            n = 0
        return self._key.read(n)

    def close(self):
        self._key.close()

    @property
    def closed(self):
        return self._key.closed

    @property
    def public_url(self):
        # Old boto versions did support never.
        # but latests seems not to https://github.com/boto/boto/blob/develop/boto/s3/connection.py#L390
        expires_in = 31536000 # 1 YEAR
        return self._key.generate_url(expires_in=expires_in, query_auth=False)


class BucketDriver(object):

    def __init__(self, bucket, prefix):
        self.bucket = bucket
        self.prefix = prefix

    def get_key(self, key_name):
        return self.bucket.get_key('%s%s' % (self.prefix, key_name))

    def new_key(self, key_name):
        return self.bucket.new_key('%s%s' % (self.prefix, key_name))

    def list_key_names(self):
        keys = self.bucket.list(prefix=self.prefix)
        return [k.name[len(self.prefix):] for k in keys]


class S3Storage(FileStorage):
    """:class:`depot.io.interfaces.FileStorage` implementation that stores files on S3.

    **This is a version implemented on boto**
    Installing ``boto`` as a dependency is required to use this.

    All the files are stored inside a bucket named ``bucket`` on ``host`` which Depot
    connects to using ``access_key_id`` and ``secret_access_key``.
 
    Additional options include:
        * ``host`` which can be used to specify an host different from Amazon
          AWS S3 Storage
        * ``policy`` which can be used to specify a canned ACL policy of either 
          ``private`` or ``public-read``.
        * ``encrypt_key`` which can be specified to use the server side
          encryption feature.
        * ``prefix`` parameter can be used to store all files under
          specified prefix. Use a prefix like **dirname/** (*see trailing slash*)
          to store in a subdirectory.
    """

    def __init__(self, access_key_id, secret_access_key, bucket=None, host=None,
                 policy=None, encrypt_key=False, prefix=''):
        warnings.warn("awss3.S3Storage is deprecated in favor of boto3.S3Storage", DeprecationWarning)
        policy = policy or CANNED_ACL_PUBLIC_READ
        assert policy in [CANNED_ACL_PUBLIC_READ, CANNED_ACL_PRIVATE], (
            "Key policy must be %s or %s" % (CANNED_ACL_PUBLIC_READ, CANNED_ACL_PRIVATE))
        self._policy = policy or CANNED_ACL_PUBLIC_READ
        self._encrypt_key = encrypt_key

        bucket_name = bucket
        if bucket_name is None:
            bucket_name = 'filedepot-%s' % (access_key_id.lower(),)

        kw = {'host': 's3.amazonaws.com'}
        if host is not None:
            kw['host'] = host

        _enable_sigv4()  # This seems to be required for publicAccessBlock to succeed.
        self._conn = S3Connection(access_key_id, secret_access_key, **kw)
        bucket = self._conn.lookup(bucket_name)
        if not bucket: 
            bucket = self._conn.create_bucket(bucket_name, headers={"x-amz-object-ownership": "BucketOwnerPreferred"})
            r = self._conn.make_request('PUT', bucket_name, 
                                        query_args="publicAccessBlock", 
                                        headers={"Content-Type": "application/xml"},
                                        data=b'<PublicAccessBlockConfiguration xmlns="http://s3.amazonaws.com/doc/2006-03-01/"><BlockPublicAcls>false</BlockPublicAcls><IgnorePublicAcls>false</IgnorePublicAcls><BlockPublicPolicy>true</BlockPublicPolicy><RestrictPublicBuckets>true</RestrictPublicBuckets></PublicAccessBlockConfiguration>')
        self._bucket_driver = BucketDriver(bucket, prefix)

    def get(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        key = self._bucket_driver.get_key(fileid)
        if key is None:
            raise IOError('File %s not existing' % fileid)

        return S3StoredFile(fileid, key)

    def __save_file(self, key, content, filename, content_type=None):
        if filename:
            filename = percent_encode(filename, safe='!#$&+-.^_`|~', encoding='utf-8')
        key.set_metadata('content-type', content_type)
        key.set_metadata(
            'x-depot-filename',
            filename
        )
        key.set_metadata('x-depot-modified', utils.timestamp())
        key.set_metadata(
            'Content-Disposition',
            make_content_disposition('inline', filename))

        if hasattr(content, 'read'):
            can_seek_and_tell = True
            try:
                pos = content.tell()
                content.seek(pos)
            except:
                can_seek_and_tell = False

            if can_seek_and_tell:
                key.set_contents_from_file(content, policy=self._policy,
                                           encrypt_key=self._encrypt_key)
            else:
                key.set_contents_from_string(content.read(), policy=self._policy,
                                             encrypt_key=self._encrypt_key)
        else:
            if isinstance(content, unicode_text):
                raise TypeError('Only bytes can be stored, not unicode')
            key.set_contents_from_string(content, policy=self._policy,
                                         encrypt_key=self._encrypt_key)

    def create(self, content, filename=None, content_type=None):
        content, filename, content_type = self.fileinfo(content, filename, content_type)
        new_file_id = str(uuid.uuid1())
        key = self._bucket_driver.new_key(new_file_id)
        self.__save_file(key, content, filename, content_type)
        return new_file_id

    def replace(self, file_or_id, content, filename=None, content_type=None):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        if isinstance(file_or_id, StoredFile) and file_or_id is content:
            # This is a backup, no need to check if file exists.
            pass
        elif not self.exists(fileid):
            # Check file existed and we are not using replace
            # as a way to force a specific file id on creation.
            raise IOError('File %s not existing' % fileid)

        content, filename, content_type = self.fileinfo(content, filename, content_type,
                                                        lambda: self.get(fileid))

        key = self._bucket_driver.get_key(fileid)
        if key is None:
            key = self._bucket_driver.new_key(fileid)

        self.__save_file(key, content, filename, content_type)
        return fileid

    def delete(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        k = self._bucket_driver.get_key(fileid)
        if k:
            k.delete()

    def exists(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        k = self._bucket_driver.get_key(fileid)
        return k is not None

    def list(self):
        return self._bucket_driver.list_key_names()


def _check_file_id(file_id):
    # Check that the given file id is valid, this also
    # prevents unsafe paths.
    try:
        uuid.UUID('{%s}' % file_id)
    except:
        raise ValueError('Invalid file id %s' % file_id)


def _enable_sigv4():
    if _enable_sigv4.DIDRUN:
        return
    
    try:
        boto.config.add_section('s3')
    except DuplicateSectionError:
        pass
    boto.config.set('s3', 'use-sigv4', 'True')
    _enable_sigv4.DIDRUN = True
_enable_sigv4.DIDRUN = False


# -*- coding: utf-8 -*-
import shutil
import unittest
import tempfile, os, cgi, base64
from PIL import Image
from sqlalchemy.exc import StatementError
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, ForeignKey
from sqlalchemy.types import Unicode, Integer, Text
from .base_sqla import setup_database, clear_database, DeclarativeBase, DBSession
from depot.fields.sqlalchemy import UploadedFileField
from depot.fields.upload import UploadedFile
from depot.io.utils import FileIntent
from depot.manager import DepotManager, get_file
from .utils import create_cgifs, OpenFiles
from depot.fields.specialized.image import UploadedImageWithThumb
from depot.fields.filters.thumbnails import WithThumbnailFilter
from depot._compat import u_, bytes_

of = OpenFiles()

def setUpModule():
    setup_database()

    DepotManager._clear()
    DepotManager.configure('default', {'depot.storage_path': './lfs'})
    DepotManager.configure('another', {'depot.storage_path': './lfs'})
    DepotManager.alias('another_alias', 'another')
    DepotManager.make_middleware(None)


def tearDownModule():
    shutil.rmtree('./lfs', ignore_errors=True)


class Document(DeclarativeBase):
    __tablename__ = 'docu'

    uid = Column(Integer, autoincrement=True, primary_key=True)
    name = Column(Unicode(16), unique=True)
    content = Column('content_col', UploadedFileField)
    photo = Column(UploadedFileField(upload_type=UploadedImageWithThumb))
    second_photo = Column(UploadedFileField(filters=(WithThumbnailFilter((12, 12), 'PNG'),)))
    targeted_content = Column(UploadedFileField(upload_storage='another_alias'))
    type = Column(Text, nullable=False, default='standard')
    directory_id = Column(Integer, ForeignKey('dir.uid'))

    __mapper_args__ = {
        'polymorphic_on': 'type'
    }


class Directory(DeclarativeBase):
    __tablename__ = 'dir'

    uid = Column(Integer, autoincrement=True, primary_key=True)
    name = Column(Unicode(16), unique=True)

    documents = relationship(Document, cascade="all, delete-orphan")


class StandardDocument(Document):
    __mapper_args__ = {'polymorphic_identity': 'standard'}


class Confidential(Document):
    __mapper_args__ = {'polymorphic_identity': 'confidential'}


class SQLATestCase(unittest.TestCase):
    FLUSH_SESSION = True

    def _session_flush(self):
        if self.FLUSH_SESSION:
            DBSession.flush()

    def file_exists(self, file_id):
        try:
            get_file(file_id)
            return True
        except IOError:
            return False


class TestSQLAAttachments(SQLATestCase):
    def setUp(self):
        self.file_content = b'this is the file content'
        self.fake_file = of.add(tempfile.NamedTemporaryFile())
        self.fake_file.write(self.file_content)
        self.fake_file.flush()
        clear_database()

    def tearDown(self):
        of.close_all()

    def test_create_fromfile(self):
        doc = Document(name=u_('Foo'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()
        DBSession.remove()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()
        assert d.content.file.read() == self.file_content
        assert d.content.file.filename == os.path.basename(self.fake_file.name)

    def test_create_polymorphic_from_file(self):
        doc = Confidential(name=u_('Secret'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()
        DBSession.remove()

        d = DBSession.query(Confidential).filter_by(name=u_('Secret')).first()
        assert d.content.file.read() == self.file_content
        assert d.content.file.filename == os.path.basename(self.fake_file.name)

    def test_edit_existing(self):
        doc = Document(name=u_('Foo2'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()
        DBSession.remove()

        d = DBSession.query(Document).filter_by(name=u_('Foo2')).first()
        old_file = d.content.path

        d.content = b'HELLO'
        new_file = d.content.path

        self._session_flush()
        DBSession.commit()
        DBSession.remove()

        assert get_file(new_file).read() == b'HELLO'

        try:
            fold = get_file(old_file)
            assert False, 'Should have raised IOError here'
        except IOError:
            pass

    def test_edit_existing_rollback(self):
        doc = Document(name=u_('Foo3'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()
        DBSession.remove()

        d = DBSession.query(Document).filter_by(name=u_('Foo3')).first()
        old_file = d.content.path

        d.content = b'HELLO'
        new_file = d.content.path

        self._session_flush()
        DBSession.rollback()
        DBSession.remove()

        assert get_file(old_file).read() == self.file_content
        assert not self.file_exists(new_file)

    def test_create_fromfield(self):
        field = create_cgifs('image/jpeg', self.fake_file, 'test.jpg')
        of.add(field.file)

        doc = Document(name=u_('Foo'), content=field)
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()
        assert d.content.file.read() == self.file_content
        assert d.content.filename == 'test.jpg'
        assert d.content.content_type == 'image/jpeg', d.content.content_type
        assert d.content.url == '/depot/%s' % d.content.path

    def test_create_empty(self):
        doc = Document(name=u_('Foo'), content=None)
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()
        assert d.content is None

    def test_delete_existing(self):
        doc = Document(name=u_('Foo2'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()
        DBSession.remove()

        d = DBSession.query(Document).filter_by(name=u_('Foo2')).first()
        old_file = d.content.path
        DBSession.delete(d)

        self._session_flush()
        DBSession.commit()
        DBSession.remove()

        assert not self.file_exists(old_file)

    def test_delete_existing_from_query(self):
        doc = Document(name=u_('Foo2'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()
        DBSession.remove()

        d = DBSession.query(Document).filter_by(name=u_('Foo2')).first()
        old_file = d.content.path

        DBSession.query(Document).filter_by(name=u_('Foo2')).delete(synchronize_session="fetch")

        self._session_flush()
        DBSession.commit()
        DBSession.remove()

        assert not self.file_exists(old_file)

    def test_delete_existing_rollback(self):
        doc = Document(name=u_('Foo3'))
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()
        DBSession.remove()

        d = DBSession.query(Document).filter_by(name=u_('Foo3')).first()
        old_file = d.content.path
        DBSession.delete(d)

        self._session_flush()
        DBSession.rollback()
        DBSession.remove()

        assert get_file(old_file).read() == self.file_content

    def test_create_with_alias(self):
        doc = Document(name=u_('Foo'))
        doc.targeted_content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()
        assert d.targeted_content.file.read() == self.file_content
        assert d.targeted_content.file.filename == os.path.basename(self.fake_file.name)
        assert d.targeted_content.depot_name == 'another'

    def test_relationship(self):
        directory = Directory(name='Parent')
        DBSession.add(directory)
        directory.documents.append(Document(name=u_('Foo'), content=of.open(self.fake_file.name, 'rb')))
        self._session_flush()
        DBSession.commit()
        
        d = DBSession.query(Directory).filter_by(name=u_('Parent')).first()
        doc = d.documents[0]
        old_file = doc.content.path
        assert self.file_exists(old_file)

        d.documents.remove(doc)
        self._session_flush()
        DBSession.commit()

        assert not self.file_exists(old_file)

    def test_relationship_rollback(self):
        directory = Directory(name='Parent')
        DBSession.add(directory)
        directory.documents.append(Document(name=u_('Foo'), content=of.open(self.fake_file.name, 'rb')))
        self._session_flush()
        DBSession.commit()
        
        d = DBSession.query(Directory).filter_by(name=u_('Parent')).first()
        doc = d.documents[0]
        old_file = doc.content.path
        assert self.file_exists(old_file)

        d.documents.remove(doc)
        self._session_flush()
        DBSession.rollback()

        assert self.file_exists(old_file)

    def test_relationship_cascade_delete(self):
        directory = Directory(name='Parent')
        DBSession.add(directory)
        directory.documents.append(Document(name=u_('Foo'), content=of.open(self.fake_file.name, 'rb')))
        self._session_flush()
        DBSession.commit()
        
        d = DBSession.query(Directory).filter_by(name=u_('Parent')).first()
        doc = d.documents[0]
        old_file = doc.content.path
        assert self.file_exists(old_file)

        DBSession.delete(d)
        self._session_flush()
        DBSession.commit()

        assert not self.file_exists(old_file)

    def test_relationship_cascade_delete_rollback(self):
        directory = Directory(name='Parent')
        DBSession.add(directory)
        directory.documents.append(Document(name=u_('Foo'), content=of.open(self.fake_file.name, 'rb')))
        self._session_flush()
        DBSession.commit()
        
        d = DBSession.query(Directory).filter_by(name=u_('Parent')).first()
        doc = d.documents[0]
        old_file = doc.content.path
        assert self.file_exists(old_file)

        DBSession.delete(d)
        self._session_flush()
        DBSession.rollback()

        assert self.file_exists(old_file)



class TestSQLAAttachmentsNoFlush(TestSQLAAttachments):
    FLUSH_SESSION = False


class TestSQLAImageAttachments(SQLATestCase):
    def setUp(self):
        self.file_content = b'''R0lGODlhEQAUAPcAAC4uLjAwMDIyMjMzMjQ0NDU1NDY2Njk2Mzg4ODo6Oj49Ozw8PD4+PkE+OEA/PkhAN0tCNk5JPFFGNV1KMFhNNVFHOFJJPVVLPVhOPXZfKXVcK2ZQNGZXNGtZMnNcNHZeNnldMHJfOn1hKXVjOH9oO0BAQEJCQkREREVFREZGRklGQ05KQ0hISEpKSkxMTE5OTlZRSlFQT19XSFBQUFJSUlRUVGFUQmFVQ2ZZQGtdQnNiQqJ/HI1uIYBnLIllKoZrK4FqLoVqL4luLIpsLpt7J515JJ50KZhzLYFnMIFlM4ZlMIFkNI1uNoJoOoVrPIlvO49yMolwPpB2O5p4Op98PaB3IKN4JqN8J6h7I6J5LaZ+LLF+ILGGG7+JG72OGLKEI7aHIrOEJL2JI7mMN76YNcGJG8SOG8WLHMONH86eEs+aFsGSG8eQHMySG9uVFduXFdeeE9eaFdScF96YE9yaFOKcEuOdEtWgFNiiEduhE96pEuqlD+qmD+KpD+yoDu6rDuysDvCuDfCvDeuwD/SzC/a2CvGwDfKxDPi5Cfi5Cvq8CeCjEuehEOagEeijEOKoEOStFMOOK8+TLM6YNNChItGgLtylKt6gMNqgON6jPOChLfi/JOSrNeGvN9KhRtykRNWkSOCnQOCpSOawQue1T+a6Su67SOGsUO/AVAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAAAAAAALAAAAAARABQAAAj+AAEIHEiwoMAACBMqXIhQgMOHDwdAfEigYsUCT6KQScNjxwGLBAyINPBhCilUmxIV6uMlw0gEMJeMGWWqFCU8hA4JEgETQYIEDcRw6qRlwgMIGvJwkfAzwYIFXQBB8qHg6VMKFawuYNBBjaIqDhhI+cGgrNmyJXooGrShBJBKcIpwiFCibl0TehDdsWBCiBxLZuIwolMGiwcTJ4gUenThBAokSVRgGFJnzhYQJ1KsyRkkhWfPK87McWPEM4sRhgItCsGitQ5PmtxYUdK6BY4rf/zwYRNmUihRpyQdaUHchQsoX/Y4amTnUqZPoG7YMO7ihfUcYNC0eRMJExMqMKweW59BfkYMEk2ykHAio3x5GvDjy58Pv4b9+/jz2w8IADs='''
        self.file_content = base64.b64decode(self.file_content)
        self.fake_file = of.add(tempfile.NamedTemporaryFile())
        self.fake_file.write(self.file_content)
        self.fake_file.flush()
        self.fake_file.seek(0)

        self.bigimage = of.add(tempfile.NamedTemporaryFile())
        blackimage = Image.frombytes('L', (1280, 1280), b"\x00" * 1280 * 1280)
        blackimage.save(self.bigimage, 'PNG')
        self.bigimage.flush()
        self.bigimage.seek(0)

        clear_database()

    def tearDown(self):
        of.close_all()

    def test_create_fromfile(self):
        doc = Document(name=u_('Foo'))
        doc.photo = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()
        assert d.photo.file.read() == self.file_content
        assert d.photo.filename == os.path.basename(self.fake_file.name)

    def test_check_assigned_type(self):
        doc = Document(name=u_('Foo'))
        doc.photo = UploadedFile(of.open(self.fake_file.name, 'rb'))
        DBSession.add(doc)

        try:
            DBSession.flush()
        except StatementError as e:
            assert 'ValueError' in str(e)
            return

        assert False, 'FLUSH did not raise exception'

    def test_create_fromfield(self):
        field = cgi.FieldStorage()
        field.filename = u_('àèìòù.gif')
        field.file = of.open(self.fake_file.name, 'rb')

        doc = Document(name=u_('Foo'), photo=field)
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()
        assert d.photo.file.read() == self.file_content
        assert d.photo.filename == field.filename
        assert d.photo.url == '/depot/%s' % d.photo.path
        assert d.photo.thumb_url == '/depot/%s' % d.photo.thumb_path
        assert d.photo.url != d.photo.thumb_url

    def test_thumbnail(self):
        field = create_cgifs('image/gif', self.fake_file, 'test.gif')
        of.add(field.file)

        doc = Document(name=u_('Foo'), photo=field)
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()
        assert os.path.exists(d.photo.thumb_file._file_path)

        thumb = Image.open(d.photo.thumb_file._file_path)
        thumb.verify()
        assert thumb.format.upper() == d.photo.thumbnail_format.upper()

    def test_maximum_size(self):
        field = create_cgifs('image/png', self.bigimage, 'test.png')
        of.add(field.file)

        doc = Document(name=u_('Foo'), photo=field)
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()

        img = Image.open(d.photo.file._file_path)
        img.verify()
        assert img.format.upper() == 'PNG'
        assert max(img.size) == 1024
        assert d.photo.filename == field.filename
        assert d.photo.content_type == 'image/png'

    def test_public_url(self):
        doc = Document(name=u_('Foo'))
        doc.photo = of.open(self.fake_file.name, 'rb')
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()

        # Hack to force object to have a public url
        d.photo._thaw()
        d.photo['_public_url'] = 'PUBLIC_URL'
        d.photo['_thumb_public_url'] = 'THUMB_PUBLIC_URL'
        d.photo._freeze()

        assert d.photo.url == 'PUBLIC_URL'
        assert d.photo.thumb_url == 'THUMB_PUBLIC_URL'

    def test_rollback(self):
        doc = Document(name=u_('Foo3'))
        doc.photo = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()

        uploaded_file = doc.photo.path
        uploaded_thumb = doc.photo.thumb_path

        DBSession.rollback()
        DBSession.remove()

        assert not self.file_exists(uploaded_file)
        assert not self.file_exists(uploaded_thumb)


class TestSQLAImageAttachmentsNoFlush(TestSQLAImageAttachments):
    FLUSH_SESSION = False


class TestSQLAThumbnailFilter(SQLATestCase):
    def setUp(self):
        self.file_content = b'''R0lGODlhEQAUAPcAAC4uLjAwMDIyMjMzMjQ0NDU1NDY2Njk2Mzg4ODo6Oj49Ozw8PD4+PkE+OEA/PkhAN0tCNk5JPFFGNV1KMFhNNVFHOFJJPVVLPVhOPXZfKXVcK2ZQNGZXNGtZMnNcNHZeNnldMHJfOn1hKXVjOH9oO0BAQEJCQkREREVFREZGRklGQ05KQ0hISEpKSkxMTE5OTlZRSlFQT19XSFBQUFJSUlRUVGFUQmFVQ2ZZQGtdQnNiQqJ/HI1uIYBnLIllKoZrK4FqLoVqL4luLIpsLpt7J515JJ50KZhzLYFnMIFlM4ZlMIFkNI1uNoJoOoVrPIlvO49yMolwPpB2O5p4Op98PaB3IKN4JqN8J6h7I6J5LaZ+LLF+ILGGG7+JG72OGLKEI7aHIrOEJL2JI7mMN76YNcGJG8SOG8WLHMONH86eEs+aFsGSG8eQHMySG9uVFduXFdeeE9eaFdScF96YE9yaFOKcEuOdEtWgFNiiEduhE96pEuqlD+qmD+KpD+yoDu6rDuysDvCuDfCvDeuwD/SzC/a2CvGwDfKxDPi5Cfi5Cvq8CeCjEuehEOagEeijEOKoEOStFMOOK8+TLM6YNNChItGgLtylKt6gMNqgON6jPOChLfi/JOSrNeGvN9KhRtykRNWkSOCnQOCpSOawQue1T+a6Su67SOGsUO/AVAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAAAAAAALAAAAAARABQAAAj+AAEIHEiwoMAACBMqXIhQgMOHDwdAfEigYsUCT6KQScNjxwGLBAyINPBhCilUmxIV6uMlw0gEMJeMGWWqFCU8hA4JEgETQYIEDcRw6qRlwgMIGvJwkfAzwYIFXQBB8qHg6VMKFawuYNBBjaIqDhhI+cGgrNmyJXooGrShBJBKcIpwiFCibl0TehDdsWBCiBxLZuIwolMGiwcTJ4gUenThBAokSVRgGFJnzhYQJ1KsyRkkhWfPK87McWPEM4sRhgItCsGitQ5PmtxYUdK6BY4rf/zwYRNmUihRpyQdaUHchQsoX/Y4amTnUqZPoG7YMO7ihfUcYNC0eRMJExMqMKweW59BfkYMEk2ykHAio3x5GvDjy58Pv4b9+/jz2w8IADs='''
        self.file_content = base64.b64decode(self.file_content)
        self.fake_file = of.add(tempfile.NamedTemporaryFile())
        self.fake_file.write(self.file_content)
        self.fake_file.flush()
        self.fake_file.seek(0)
        clear_database()

    def tearDown(self):
        of.close_all()

    def test_create_fromfile(self):
        doc = Document(name=u_('Foo'))
        doc.second_photo = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()
        assert d.second_photo.file.read() == self.file_content
        assert d.second_photo.filename == os.path.basename(self.fake_file.name)

    def test_create_from_bytes(self):
        doc = Document(name=u_('Foo'))
        doc.second_photo = self.file_content
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()
        assert d.second_photo.file.read() == self.file_content
        assert d.second_photo.filename == 'unnamed'

    def test_create_fromfield(self):
        field = cgi.FieldStorage()
        field.filename = u_('àèìòù.gif')
        field.file = of.open(self.fake_file.name, 'rb')

        doc = Document(name=u_('Foo'), second_photo=field)
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()
        assert d.second_photo.file.read() == self.file_content
        assert d.second_photo.filename == field.filename
        assert d.second_photo.url == '/depot/%s' % d.second_photo.path
        assert d.second_photo.thumb_12x12_url == '/depot/%s' % d.second_photo.thumb_12x12_path
        assert d.second_photo.url != d.second_photo.thumb_12x12_url

    def test_create_fileintent(self):
        field = FileIntent(of.open(self.fake_file.name, 'rb'), u_('àèìòù.gif'), 'image/gif')

        doc = Document(name=u_('Foo'), second_photo=field)
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()
        assert d.second_photo.file.read() == self.file_content
        assert d.second_photo.filename == field._filename
        assert d.second_photo.url == '/depot/%s' % d.second_photo.path
        assert d.second_photo.thumb_12x12_url == '/depot/%s' % d.second_photo.thumb_12x12_path
        assert d.second_photo.url != d.second_photo.thumb_12x12_url
        assert d.second_photo.content_type == field._content_type

    def test_thumbnail(self):
        field = create_cgifs('image/gif', self.fake_file, 'test.gif')
        of.add(field.file)

        doc = Document(name=u_('Foo'), second_photo=field)
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()

        thumbnail_local_path = DepotManager.get_file(d.second_photo.thumb_12x12_path)._file_path
        assert os.path.exists(thumbnail_local_path)

        thumb = Image.open(thumbnail_local_path)
        thumb.verify()
        assert thumb.format.upper() == 'PNG'
        assert max(thumb.size) == 12

    def test_public_url(self):
        doc = Document(name=u_('Foo'))
        doc.second_photo = of.open(self.fake_file.name, 'rb')
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()
        DBSession.commit()

        d = DBSession.query(Document).filter_by(name=u_('Foo')).first()

        # Hack to force object to have a public url
        d.second_photo._thaw()
        d.second_photo['_public_url'] = 'PUBLIC_URL'
        d.second_photo._freeze()

        assert d.second_photo.url == 'PUBLIC_URL'
        assert d.second_photo.thumb_12x12_url.startswith('/depot/default/')

    def test_rollback(self):
        doc = Document(name=u_('Foo3'))
        doc.second_photo = of.open(self.fake_file.name, 'rb')
        DBSession.add(doc)
        self._session_flush()

        uploaded_file = doc.second_photo.path
        uploaded_thumb = doc.second_photo.thumb_12x12_path

        DBSession.rollback()
        DBSession.remove()

        assert not self.file_exists(uploaded_file)
        assert not self.file_exists(uploaded_thumb)


class TestSQLAThumbnailFilterNoFlush(TestSQLAThumbnailFilter):
    FLUSH_SESSION = False



# -*- coding: utf-8 -*-
import unittest
import os
import shutil
import json
import uuid
from urllib.parse import parse_qs
from depot.middleware import _FileIter
from depot.manager import DepotManager
from webtest import TestApp
from depot._compat import u_, unquote, PY2


FILE_CONTENT = b'HELLO WORLD'


class RootController:
    UPLOADED_FILES = []

    def index(self):
        return dict(files=self.UPLOADED_FILES)

    def clear_files(self):
        self.UPLOADED_FILES = []
        return dict(files=self.UPLOADED_FILES)

    def depotskipped(self):
        return dict(ok=True)

    def depot(self):
        # this should never be called
        return dict(ok=False)

    def create_file(self, lang=('en',)):
        lang = lang[0]
        fname = {'en': 'hello.txt',
                 'ru': u_('Крупный'),
                 'it': u_('àèìòù')}.get(lang, 'unknown')

        self.UPLOADED_FILES += [DepotManager.get().create(FILE_CONTENT,
                                                          filename=fname)]
        return dict(files=self.UPLOADED_FILES,
                    uploaded_to=DepotManager.get_default(),
                    last=self.UPLOADED_FILES[-1])


class WSGIApplication:
    def __init__(self, controller):
        self._controller = controller

    def __call__(self, env, start_response):
        path = env['PATH_INFO'].strip('/') or 'index'
        try:
            action = getattr(self._controller, path)
        except AttributeError:
            start_response('404 Not Found', [('Content-Type','text/plain')])
            return [b'Not Found']
        if env['REQUEST_METHOD'] == 'POST':
            query_string = env['wsgi.input'].read().decode('utf-8')
        else:
            query_string = env['QUERY_STRING']
        resp = json.dumps(action(**parse_qs(query_string))).encode('utf-8')
        start_response('200 OK', [('Content-Type','application/json')])
        return [resp]


class BaseWSGITests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.wsgi_app = WSGIApplication(RootController())

    def make_app(self, **options):
        wsgi_app = DepotManager.make_middleware(self.wsgi_app, **options)
        return TestApp(wsgi_app)



class TestWSGIMiddleware(BaseWSGITests):
    def setUp(self):
        DepotManager._clear()
        DepotManager.configure('default', {'depot.storage_path': './lfs'})

    def tearDown(cls):
        shutil.rmtree('./lfs', ignore_errors=True)

    def test_invalid_mountpoint(self):
        try:
            DepotManager.make_middleware(self.wsgi_app, mountpoint='hello')
        except ValueError as err:
            assert 'mountpoint must be an absolute path' in str(err)
        else:
            assert False, 'Should have raised ValueError'
    
    def test_subdirectory_mountpoint(self):
        wsgi_app = DepotManager.make_middleware(self.wsgi_app, mountpoint='/hello/lfs')
        app = TestApp(wsgi_app)
        new_file = app.post('/create_file').json
        uploaded_file = app.get(DepotManager.url_for('%(uploaded_to)s/%(last)s' % new_file))
        assert uploaded_file.body == FILE_CONTENT

    def test_serving_files(self):
        app = self.make_app()
        new_file = app.post('/create_file').json

        uploaded_file = app.get(DepotManager.url_for('%(uploaded_to)s/%(last)s' % new_file))
        assert uploaded_file.body == FILE_CONTENT

    def test_headers_are_there(self):
        app = self.make_app()
        new_file = app.post('/create_file').json

        uploaded_file = app.get(DepotManager.url_for('%(uploaded_to)s/%(last)s' % new_file))
        assert uploaded_file.headers['Content-Type'] == 'text/plain'
        assert 'ETag' in uploaded_file.headers
        assert 'Last-Modified' in uploaded_file.headers
        assert 'Expires' in uploaded_file.headers

    def test_caching_unmodified(self):
        app = self.make_app()
        new_file = app.post('/create_file').json

        uploaded_file = app.get(DepotManager.url_for('%(uploaded_to)s/%(last)s' % new_file))
        last_modified = uploaded_file.headers['Last-Modified']

        unmodified_file = app.get('/depot/default/%(last)s' % new_file,
                                  headers=[('If-Modified-Since', last_modified)],
                                  status=304)
        assert 'Not Modified' in unmodified_file.status, unmodified_file

    def test_caching_etag(self):
        app = self.make_app()
        new_file = app.post('/create_file').json

        uploaded_file = app.get(DepotManager.url_for('%(uploaded_to)s/%(last)s' % new_file))
        etag = uploaded_file.headers['ETag']

        unmodified_file = app.get('/depot/default/%(last)s' % new_file,
                                  headers=[('If-None-Match', etag)],
                                  status=304)
        assert 'Not Modified' in unmodified_file.status, unmodified_file

    def test_post_is_forwarded_to_app(self):
        app = self.make_app()
        new_file = app.post('/create_file').json

        file_path = DepotManager.url_for('%(uploaded_to)s/%(last)s' % new_file)
        uploaded_file = app.post(file_path, status=404)
        assert 'Not Found' in uploaded_file.status, uploaded_file

    def test_forwards_to_app(self):
        app = self.make_app()
        new_file = app.post('/create_file').json

        files = app.get('/').json
        assert new_file['last'] in files['files'], (new_file, files)

    def test_forwards_to_app_begins_with_endpoint(self):
        app = self.make_app()

        resp = app.get('/depotskipped').json
        assert resp['ok'] == True

    def test_404_on_nofile(self):
        app = self.make_app()

        missing = app.get('/depot', status=404)
        assert 'Not Found' in missing.status

    def test_404_on_missing_file(self):
        app = self.make_app()
        missing = app.get('/depot/default/hello', status=404)
        assert 'Not Found' in missing.status
        missing = app.get('/depot/nodepot', status=404)
        assert 'Not Found' in missing.status
        missing = app.get('/depot/nodepot/hello', status=404)
        assert 'Not Found' in missing.status

    def test_invalid_unmodified_header(self):
        app = self.make_app()
        new_file = app.post('/create_file').json

        uploaded_file = app.get(DepotManager.url_for('%(uploaded_to)s/%(last)s' % new_file))

        unmodified_file = app.get('/depot/default/%(last)s' % new_file,
                                  headers=[('If-Modified-Since', 'HELLO WORLD')],
                                  status=400)
        assert 'Bad Request' in unmodified_file.status, unmodified_file

    def test_serving_files_with_wsgifilewrapper(self):
        app = self.make_app(replace_wsgi_filewrapper=True)
        new_file = app.post('/create_file').json

        uploaded_file = app.get(DepotManager.url_for('%(uploaded_to)s/%(last)s' % new_file))
        assert uploaded_file.body == FILE_CONTENT
        assert uploaded_file.request.environ['wsgi.file_wrapper'] is _FileIter

    def test_serving_files_content_disposition(self):
        app = self.make_app()
        new_file = app.post('/create_file', params={'lang': 'ru'}).json

        uploaded_file = app.get(DepotManager.url_for('%(uploaded_to)s/%(last)s' % new_file))
        content_disposition = uploaded_file.headers['Content-Disposition']

        if PY2:
            assert content_disposition == "inline;filename=\"unknown\";filename*=utf-8''%D0%9A%D1%80%D1%83%D0%BF%D0%BD%D1%8B%D0%B9", content_disposition
        else:
            assert content_disposition == "inline;filename=\"Krupnyy\";filename*=utf-8''%D0%9A%D1%80%D1%83%D0%BF%D0%BD%D1%8B%D0%B9", content_disposition


        new_file = app.post('/create_file', params={'lang': 'it'}).json
        uploaded_file = app.get(DepotManager.url_for('%(uploaded_to)s/%(last)s' % new_file))
        content_disposition = uploaded_file.headers['Content-Disposition']
        _, asciiname, uniname = content_disposition.split(';')
        assert asciiname == 'filename="aeiou"', asciiname
        assert u_(unquote(uniname[17:])) == u_('àèìòù'), unquote(uniname[17:])


class TestS3TestWSGIMiddleware(BaseWSGITests):
    def setUp(self):
        DepotManager._clear()

        try:
            global S3Storage
            from depot.io.awss3 import S3Storage
        except ImportError:
            self.skipTest('Boto not installed')

        env = os.environ
        access_key_id = env.get('AWS_ACCESS_KEY_ID')
        secret_access_key = env.get('AWS_SECRET_ACCESS_KEY')
        if access_key_id is None or secret_access_key is None:
            self.skipTest('Amazon S3 credentials not available')

        PID = os.getpid()
        NODE = str(uuid.uuid1()).rsplit('-', 1)[-1]  # Travis runs multiple tests concurrently
        default_bucket_name = 'filedepot-%s' % (access_key_id.lower(), )
        cred = (access_key_id, secret_access_key)
        bucket_name = 'filedepot-testfs-%s-%s-%s' % (access_key_id.lower(), NODE, PID)

        DepotManager.configure('awss3', {'depot.backend': 'depot.io.awss3.S3Storage',
                                         'depot.access_key_id': access_key_id,
                                         'depot.secret_access_key': secret_access_key,
                                         'depot.bucket': bucket_name})
        DepotManager.set_default('awss3')

    def tearDown(self):
        store = DepotManager.get('awss3')
        if not store._conn.lookup(store._bucket_driver.bucket.name):
            return
        
        keys = [key.name for key in store._bucket_driver.bucket]
        if keys:
            store._bucket_driver.bucket.delete_keys(keys)

        try:
            store._conn.delete_bucket(store._bucket_driver.bucket.name)
            while store._conn.lookup(store._bucket_driver.bucket.name):
                # Wait for bucket to be deleted, to avoid flaky tests...
                time.sleep(0.5)
        except:
            pass
    
    def test_public_url_gets_redirect(self):
        app = self.make_app()
        new_file = app.post('/create_file').json

        file_path = DepotManager.url_for('%(uploaded_to)s/%(last)s' % new_file)
        uploaded_file = app.get(file_path, status=301)
        location = uploaded_file.headers['Location']
        assert 'https://filedepot-testfs-' in location
        assert 's3.amazonaws.com' in location



from abc import ABCMeta, abstractmethod
from depot._compat import with_metaclass
from depot.manager import DepotManager


class FileFilter(with_metaclass(ABCMeta, object)):
    """Interface that must be implemented by file filters.

    File filters get executed whenever a file is stored on the database
    using one of the supported fields. Can be used to add additional data
    to the stored file or change it. When file filters are run the file
    has already been stored.

    """
    @abstractmethod
    def on_save(self, uploaded_file):  # pragma: no cover
        """Filters are required to provide their own implementation"""
        return


class DepotFileInfo(with_metaclass(ABCMeta, dict)):
    """Keeps information on a content related to a specific depot.

    By itself the DepotFileInfo does nothing, it is required to implement
    a :meth:`process_content` method that actually saves inside the
    file info the information related to the content. The only information
    which is saved by default is the depot name itself.

    It is a specialized dictionary that provides also attribute style access,
    the dictionary parent permits easy encoding/decoding to most marshalling
    systems.

    """
    def __init__(self, content, depot_name=None):
        super(DepotFileInfo, self).__init__()
        self._thaw()

        if isinstance(content, dict):
            object.__setattr__(self, 'original_content', None)
            self.update(content)
        else:
            object.__setattr__(self, 'original_content', content)
            if depot_name is None:
                depot_name = DepotManager.get_default()

            depot_name = DepotManager.resolve_alias(depot_name)
            if not depot_name:
                raise ValueError('Storage has not been found in DEPOT')

            self['depot_name'] = depot_name
            self['files'] = []
            self.process_content(content)

        self._freeze()

    @abstractmethod
    def process_content(self, content, filename=None, content_type=None):  # pragma: no cover
        """Process content in the given depot.

        This is implemented by subclasses to provide some kind of behaviour on the
        content in the related Depot. The default implementation is provided by
        :class:`depot.fields.upload.UploadedFile` which stores the content into
        the depot.

        """
        return

    def __getitem__(self, key):
        return dict.__getitem__(self, key)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setitem__(self, key, value):
        if object.__getattribute__(self, '_frozen'):
            raise TypeError('Already saved files are immutable')
        return dict.__setitem__(self, key, value)

    __setattr__ = __setitem__

    def __delattr__(self, name):
        if object.__getattribute__(self, '_frozen'):
            raise TypeError('Already saved files are immutable')

        try:
            del self[name]
        except KeyError:
            raise AttributeError(name)

    def __delitem__(self, key):
        if object.__getattribute__(self, '_frozen'):
            raise TypeError('Already saved files are immutable')
        dict.__delitem__(self, key)

    def _apply_filters(self, filters):
        if self.original_content is not None:
            self._thaw()
            for filt in filters:
                filt.on_save(self)
            self._freeze()

    def _freeze(self):
        object.__setattr__(self, '_frozen', True)

    def _thaw(self):
        object.__setattr__(self, '_frozen', False)


# -*- coding: utf-8 -*-
import uuid
import unittest
import json
from flaky import flaky
import shutil
import mock
import datetime
from io import BytesIO
from tempfile import NamedTemporaryFile
import os
from depot.io.utils import FileIntent
from tests.utils import create_cgifs

FILE_CONTENT = b'HELLO WORLD'


class BaseStorageTestFixture(object):
    def test_creation_readback(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        f = self.fs.get(file_id)
        assert f.read() == FILE_CONTENT

    def test_creation_metadata(self):
        with mock.patch('depot.io.utils.timestamp', return_value='2001-01-01 00:00:01'):
            file_id = self.fs.create(FILE_CONTENT, 'file.txt')

        f = self.fs.get(file_id)
        assert f.filename == 'file.txt', f.filename
        assert f.last_modified == datetime.datetime(2001, 1, 1, 0, 0, 1), f.last_modified
        assert f.content_type == 'text/plain', f.content_type

    def test_notexisting(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        assert self.fs.exists(file_id)
        self.fs.delete(file_id)
        assert not self.fs.exists(file_id)

        with self.assertRaises(IOError):
            self.fs.get(file_id)

    def test_invalidid(self):
        with self.assertRaises(ValueError):
            f = self.fs.get('NOTANID')

    def test_creation_inputs(self):
        temp = NamedTemporaryFile()
        temp.write(FILE_CONTENT)
        temp.seek(0)

        for d in (FILE_CONTENT,
                  BytesIO(FILE_CONTENT),
                  temp,
                  create_cgifs('text/plain', FILE_CONTENT, 'file.txt'),
                  FileIntent(FILE_CONTENT, 'file.txt', 'text/plain')):
            fid = self.fs.create(d, 'filename')
            f = self.fs.get(fid)
            assert f.read() == FILE_CONTENT

    def test_content_type_by_name(self):
        for fname, ctype in (('filename', 'application/octet-stream'),
                             ('image.png', 'image/png'),
                             ('file.txt', 'text/plain')):
            file_id = self.fs.create(FILE_CONTENT, fname)
            f = self.fs.get(file_id)
            assert f.content_type == ctype, (fname, ctype)

    def test_cgifieldstorage(self):
        cgifs = create_cgifs('text/plain', FILE_CONTENT, 'file.txt')
        file_id = self.fs.create(cgifs)

        f = self.fs.get(file_id)
        assert f.content_type == 'text/plain'
        assert f.filename == 'file.txt'
        assert f.read() == FILE_CONTENT

    def test_fileintent(self):
        file_id = self.fs.create(FileIntent(FILE_CONTENT, 'file.txt', 'text/plain'))

        f = self.fs.get(file_id)
        assert f.content_type == 'text/plain', f.content_type
        assert f.filename == 'file.txt', f.filename
        assert f.read() == FILE_CONTENT

    def test_filewithname(self):
        temp = NamedTemporaryFile()
        temp.write(FILE_CONTENT)
        temp.seek(0)

        file_id = self.fs.create(temp)

        f = self.fs.get(file_id)
        assert f.content_type == 'application/octet-stream'
        assert temp.name.endswith(f.filename)
        assert f.read() == FILE_CONTENT

    def test_filewithnonasciiname(self):
        filename = u'些公.pdf'
        temp = NamedTemporaryFile()
        temp.write(FILE_CONTENT)
        temp.seek(0)

        file_id = self.fs.create(
            temp,
            filename=filename,
            content_type='application/pdf'
        )

        f = self.fs.get(file_id)
        assert f.content_type == 'application/pdf'
        assert f.read() == FILE_CONTENT

    def test_another_copy(self):
        file_id = self.fs.create(FILE_CONTENT, filename='file.txt', content_type='text/plain')
        f = self.fs.get(file_id)

        file2_id = self.fs.create(f)
        assert file2_id != f.file_id

        f2 = self.fs.get(file2_id)
        assert f2.filename == f.filename
        assert f2.content_type == f.content_type
        assert f2.filename == 'file.txt', (f.filename, f2.filename)
        assert f2.content_type == 'text/plain'

    def test_copy_to_other_storage(self):
        file_id = self.fs.create(FILE_CONTENT, filename='file.txt', content_type='text/plain')
        f = self.fs.get(file_id)

        other_storage = self.get_storage("otherbucket-%s" % uuid.uuid1().hex)
        try:
            other_storage.replace(f, f)

            assert other_storage.exists(f.file_id)
            assert other_storage.get(f.file_id).read() == FILE_CONTENT
            assert other_storage.get(f.file_id).filename == f.filename
        finally:
            self.delete_storage(other_storage)

    def test_backup(self):
        file_ids = set()
        for i in range(10):
            file_ids.add(
                self.fs.create(FILE_CONTENT, filename='file-%s.txt' %i,
                               content_type='text/plain')
            )

        other_storage = self.get_storage("otherbucket-%s" % uuid.uuid1().hex)
        existing_files = self.fs.list()
        try:
            for f_id in existing_files:
                f = self.fs.get(f_id)
                other_storage.replace(f, f)

            assert set(other_storage.list()) == file_ids
        finally:
            self.delete_storage(other_storage)


    def test_repr(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        f = self.fs.get(file_id)

        assert repr(f).endswith('%s filename=file.txt '
                                'content_type=text/plain '
                                'last_modified=%s>' % (f.file_id, f.last_modified))

    def test_replace_only_existing(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        assert self.fs.exists(file_id)
        self.fs.delete(file_id)
        assert not self.fs.exists(file_id)

        with self.assertRaises(IOError):
            self.fs.replace(file_id, FILE_CONTENT)

    def test_replace_invalidid(self):
        with self.assertRaises(ValueError):
            self.fs.replace('INVALIDID', FILE_CONTENT)

    def test_replace_keeps_filename(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        f = self.fs.get(file_id)

        self.fs.replace(f, b'NEW CONTENT')
        f2 = self.fs.get(f.file_id)

        assert f2.file_id == f.file_id
        assert f.filename == f2.filename
        assert f.read() == b'NEW CONTENT'
        assert f2.content_type == 'text/plain'

    def test_replace_updates_data(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        f = self.fs.get(file_id)

        self.fs.replace(f, b'NEW CONTENT', 'file.png')
        f2 = self.fs.get(f.file_id)

        assert f2.file_id == f.file_id
        assert f2.filename == 'file.png'
        assert f2.read() == b'NEW CONTENT'
        assert f2.content_type == 'image/png'

    def test_exists(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        assert self.fs.exists(file_id)

        self.fs.delete(file_id)
        assert not self.fs.exists(file_id)

    def test_list(self):
        file_ids = list()
        for i in range(3):
            file_ids.append(self.fs.create(FILE_CONTENT, 'file_for_listing-{0}.txt'.format(i)))

        existing_files = self.fs.list()
        for _id in file_ids:
            assert _id in existing_files, ("{0} not in {1}".format(_id, file_ids))

    def test_exists_invalidid(self):
        with self.assertRaises(ValueError):
            self.fs.exists('INVALIDID')

    def test_delete(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        assert self.fs.exists(file_id)

        self.fs.delete(file_id)
        assert not self.fs.exists(file_id)

    def test_delete_idempotent(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        assert self.fs.exists(file_id)

        self.fs.delete(file_id)
        self.fs.delete(file_id)
        assert not self.fs.exists(file_id)

    def test_delete_invalidid(self):
        with self.assertRaises(ValueError):
            self.fs.delete('INVALIDID')

    def test_stored_files_are_only_readable(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        f = self.fs.get(file_id)
        assert f.readable()
        assert not f.writable()
        assert not f.seekable()

    def test_storing_unicode_is_prevented(self):
        with self.assertRaises(TypeError):
            file_id = self.fs.create(FILE_CONTENT.decode('utf-8'), 'file.txt')

    def test_replacing_unicode_is_prevented(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        f = self.fs.get(file_id)

        with self.assertRaises(TypeError):
            self.fs.replace(f, FILE_CONTENT.decode('utf-8'))

    def test_closing_file(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        f = self.fs.get(file_id)

        assert f.closed is False
        with f:
            f.read()
        assert f.closed is True

    def test_reading_from_closed_file_is_prevented(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        f = self.fs.get(file_id)
        f.close()

        with self.assertRaises(ValueError):
            f.read()

    def test_name_is_an_alias_for_filename(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        f = self.fs.get(file_id)
        assert f.name == f.filename

    def test_create_from_bytes_with_default_filename(self):
        file_id = self.fs.create(b'this is the content')
        f = self.fs.get(file_id)
        assert f.name == 'unnamed'


class TestLocalFileStorage(unittest.TestCase, BaseStorageTestFixture):
    @classmethod
    def get_storage(cls, bucket_name):
        from depot.io.local import LocalFileStorage
        return LocalFileStorage('./lfs/%s' % bucket_name)

    @classmethod
    def delete_storage(cls, storage):
        shutil.rmtree(storage.storage_path, ignore_errors=True)

    def setUp(self):
        self.fs = self.get_storage("default_bucket")

    def tearDown(self):
        self.delete_storage(self.fs)


class TestGridFSFileStorage(unittest.TestCase, BaseStorageTestFixture):
    @classmethod
    def get_storage(cls, collection_name):
        from depot.io.gridfs import GridFSStorage
        import pymongo.errors
        try:
            fs = GridFSStorage('mongodb://localhost/gridfs_example?serverSelectionTimeoutMS=100', collection_name)
            fs._gridfs.exists("")  # Any operation to test that mongodb is up.
            return fs
        except pymongo.errors.ConnectionFailure:
            return None

    @classmethod
    def delete_storage(cls, storage):
        storage._db.drop_collection('%s.files' % storage._collection)
        storage._db.drop_collection('%s.chunks' % storage._collection)

    @classmethod
    def setUpClass(cls):
        try:
            from depot.io.gridfs import GridFSStorage
        except ImportError:
            cls.skipTest('PyMongo not installed')

        cls.fs = cls.get_storage('testfs')

    def setUp(self):
        if self.fs is None:
            self.skipTest('MongoDB not running')

    def tearDown(self):
        self.delete_storage(self.fs)


@flaky
class TestS3FileStorage(unittest.TestCase, BaseStorageTestFixture):

    @classmethod
    def get_storage(cls, bucket_name):
        from depot.io.awss3 import S3Storage

        env = os.environ
        access_key_id = env.get('AWS_ACCESS_KEY_ID')
        secret_access_key = env.get('AWS_SECRET_ACCESS_KEY')
        if access_key_id is None or secret_access_key is None:
            return None
        
        return S3Storage(access_key_id, secret_access_key, bucket_name)

    @classmethod
    def delete_storage(cls, storage):
        keys = [key.name for key in storage._bucket_driver.bucket]
        if keys:
            storage._bucket_driver.bucket.delete_keys(keys)

        try:
            storage._conn.delete_bucket(storage._bucket_driver.bucket.name)
        except:
            pass

    @classmethod
    def setUpClass(cls):
        try:
            from depot.io.awss3 import S3Storage
        except ImportError:
            raise unittest.SkipTest('Boto not installed')

        BUCKET_NAME = 'fdtest-%s-%s' % (uuid.uuid1(), os.getpid())
        cls.fs = cls.get_storage(BUCKET_NAME)
        if cls.fs is None:
            raise unittest.SkipTest('Amazon S3 credentials not available')

    def tearDown(self):
        keys = [key.name for key in self.fs._bucket_driver.bucket]
        if keys:
            self.fs._bucket_driver.bucket.delete_keys(keys)

    @classmethod
    def tearDownClass(cls):
        cls.delete_storage(cls.fs)


@flaky
class TestS3FileStorageWithPrefix(TestS3FileStorage):

    @classmethod
    def get_storage(cls, bucket_name):
        from depot.io.awss3 import S3Storage

        env = os.environ
        access_key_id = env.get('AWS_ACCESS_KEY_ID')
        secret_access_key = env.get('AWS_SECRET_ACCESS_KEY')
        if access_key_id is None or secret_access_key is None:
            return None

        return S3Storage(
            access_key_id,
            secret_access_key,
            bucket_name,
            prefix='my-prefix/')


class TestMemoryFileStorage(unittest.TestCase, BaseStorageTestFixture):
    @classmethod
    def get_storage(cls, _):
        from depot.io.memory import MemoryFileStorage
        return MemoryFileStorage()

    @classmethod
    def delete_storage(cls, storage):
        map(storage.delete, storage.list())

    def setUp(self):
        self.fs = self.get_storage(None)

    def tearDown(self):
        self.delete_storage(self.fs)


@flaky
class TestBoto3FileStorage(unittest.TestCase, BaseStorageTestFixture):

    @classmethod
    def get_storage(cls, bucket_name):
        from depot.io.boto3 import S3Storage

        env = os.environ
        access_key_id = env.get('AWS_ACCESS_KEY_ID')
        secret_access_key = env.get('AWS_SECRET_ACCESS_KEY')
        if access_key_id is None or secret_access_key is None:
            return None

        return S3Storage(access_key_id, secret_access_key, bucket_name)

    @classmethod
    def delete_storage(cls, storage):
        for obj in storage._bucket_driver.bucket.objects.all():
            obj.delete()
        try:
            storage._bucket_driver.bucket.delete()
        except:
            pass

    @classmethod
    def setUpClass(cls):
        try:
            from depot.io.boto3 import S3Storage
        except ImportError:
            raise unittest.SkipTest('Boto3 not installed')

        BUCKET_NAME = 'fdtest-%s-%s' % (uuid.uuid1(), os.getpid())
        cls.fs = cls.get_storage(BUCKET_NAME)
        if not cls.fs:
            raise unittest.SkipTest('Amazon S3 credentials not available')

    def tearDown(self):
        for obj in self.fs._bucket_driver.bucket.objects.all():
            obj.delete()

    @classmethod
    def tearDownClass(cls):
        cls.delete_storage(cls.fs)


@flaky
class TestGCSFileStorage(unittest.TestCase, BaseStorageTestFixture):

    @classmethod
    def get_storage(cls, bucket_name):
        from depot.io.gcs import GCSStorage

        env = os.environ
        google_credentials = env.get('GOOGLE_SERVICE_CREDENTIALS')
        if not google_credentials:
            return None
        google_credentials = json.loads(google_credentials)

        return GCSStorage(project_id=google_credentials['project_id'],
                          credentials=google_credentials, 
                          bucket=bucket_name)

    @classmethod
    def delete_storage(cls, storage):
        for blob in storage.bucket.list_blobs():
            blob.delete()
        try:
            storage.bucket.delete()
        except:
            pass

    @classmethod
    def setUpClass(cls):
        try:
            from depot.io.gcs import GCSStorage
        except ImportError:
            raise unittest.SkipTest('Google Cloud Storage not installed')

        BUCKET_NAME = 'fdtest-%s-%s' % (uuid.uuid1(), os.getpid())
        cls.fs = cls.get_storage(bucket_name=BUCKET_NAME)
        if cls.fs is None:
            raise unittest.SkipTest('GOOGLE_SERVICE_CREDENTIALS environment variable not set')

    def tearDown(self):
        for blob in self.fs.bucket.list_blobs():
            blob.delete()

    @classmethod
    def tearDownClass(cls):
        cls.delete_storage(cls.fs)


# -*- coding: utf-8 -*-
import os
import uuid
import unittest
import mock
import requests
from flaky import flaky
from unittest import SkipTest

from depot._compat import PY2, unicode_text


S3Storage = None
FILE_CONTENT = b'HELLO WORLD'


@flaky
class TestS3FileStorage(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # Travis runs multiple tests concurrently on fake machines that might
        # collide on pid and hostid, so use an uuid1 which should be fairly random
        # thanks to clock_seq
        self.run_id = '%s-%s' % (uuid.uuid1().hex, os.getpid())

        try:
            global S3Storage
            from depot.io.boto3 import S3Storage
        except ImportError:
            raise SkipTest('Boto not installed')
        
        from botocore.exceptions import ClientError
        env = os.environ
        access_key_id = env.get('AWS_ACCESS_KEY_ID')
        secret_access_key = env.get('AWS_SECRET_ACCESS_KEY')
        if access_key_id is None or secret_access_key is None:
           raise SkipTest('Amazon S3 credentials not available')

        self.default_bucket_name = 'filedepot-%s' % (access_key_id.lower(), )
        self.cred = (access_key_id, secret_access_key)
        self.bucket = 'filedepot-testfs-%s' % self.run_id
        self.fs = S3Storage(*self.cred, bucket=self.bucket)

    @classmethod
    def tearDownClass(self):
        buckets = set(
            b['Name'] for b in self.fs._bucket_driver.s3.meta.client.list_buckets()['Buckets']
        )
        if self.fs._bucket_driver.bucket.name not in buckets:
            # Bucket wasn't created, probably due to monkey patching, just skip.
            return

        for obj in self.fs._bucket_driver.bucket.objects.all():
            obj.delete()

        try:
            self.fs._bucket_driver.bucket.delete()
        except:
            pass
        else:
            self.fs._bucket_driver.bucket.wait_until_not_exists()

    def test_fileoutside_depot(self):
        fid = str(uuid.uuid1())
        key = self.fs._bucket_driver.new_key(fid)
        key.put(Body=FILE_CONTENT)

        f = self.fs.get(fid)
        assert f.read() == FILE_CONTENT

    def test_creates_bucket_when_missing(self):
        created_buckets = []
        def mock_make_api_call(_, operation_name, kwarg):
            if operation_name == 'ListBuckets':
                return {'Buckets': []}
            elif operation_name == 'CreateBucket':
                created_buckets.append(kwarg['Bucket'])
                return None
            elif operation_name == 'HeadBucket':
                if kwarg['Bucket'] in created_buckets:
                    return {'ResponseMetadata': {'HTTPStatusCode': 200}}
                else:
                    return {}
            else:
                assert False, 'Unexpected Call'

        from depot.io.boto3 import CANNED_ACL_PRIVATE
        with mock.patch('botocore.client.BaseClient._make_api_call', new=mock_make_api_call):
            S3Storage(*self.cred, policy=CANNED_ACL_PRIVATE)
        assert created_buckets == [self.default_bucket_name]

    def test_bucket_failure(self):
        from botocore.exceptions import ClientError
        def mock_make_api_call(_, operation_name, kwarg):
            if operation_name == 'ListBuckets':
                raise ClientError(error_response={'Error': {'Code': 500}},
                                  operation_name=operation_name)

        try:
            with mock.patch('botocore.client.BaseClient._make_api_call', new=mock_make_api_call):
                S3Storage(*self.cred)
        except ClientError:
            pass
        else:
            assert False, 'Should have reraised ClientError'

    def test_client_receives_extra_args(self):
        with mock.patch('boto3.session.Session.client'), mock.patch('boto3.session.Session.resource') as mockresource:
            S3Storage(*self.cred, endpoint_url='http://somehwere.it', region_name='worlwide')
        mockresource.assert_called_once_with('s3', endpoint_url='http://somehwere.it',
                                             region_name='worlwide')

    def test_get_key_failure(self):
        from botocore.exceptions import ClientError

        from botocore.client import BaseClient
        make_api_call = BaseClient._make_api_call
        def mock_make_api_call(cli, operation_name, kwarg):
            if operation_name == 'HeadObject':
                raise ClientError(error_response={'Error': {'Code': 500}},
                                  operation_name=operation_name)
            return make_api_call(cli, operation_name, kwarg)

        fs = S3Storage(*self.cred)
        try:
            with mock.patch('botocore.client.BaseClient._make_api_call', new=mock_make_api_call):
                fs.get(uuid.uuid1())
        except ClientError:
            pass
        else:
            assert False, 'Should have reraised ClientError'

    def test_invalid_modified(self):
        fid = str(uuid.uuid1())
        key = self.fs._bucket_driver.new_key(fid)
        key.put(Body=FILE_CONTENT, Metadata={'x-depot-modified': 'INVALID'})

        f = self.fs.get(fid)
        assert f.last_modified is None, f.last_modified

    def test_public_url(self):
        fid = str(uuid.uuid1())
        key = self.fs._bucket_driver.new_key(fid)
        key.put(Body=FILE_CONTENT)

        f = self.fs.get(fid)
        assert '.s3.amazonaws.com' in f.public_url, f.public_url
        assert f.public_url.endswith('/%s' % fid), f.public_url

    def test_content_disposition(self):
        file_id = self.fs.create(b'content', unicode_text('test.txt'), 'text/plain')
        test_file = self.fs.get(file_id)
        response = requests.get(test_file.public_url)
        assert response.headers['Content-Disposition'] == "inline;filename=\"test.txt\";filename*=utf-8''test.txt"

    def test_storage_class(self):
        fs_ia = S3Storage(*self.cred, bucket=self.bucket, storage_class='STANDARD_IA')
        fid = fs_ia.create(FILE_CONTENT)

        key = self.fs._bucket_driver.get_key(fid)
        assert key.storage_class == 'STANDARD_IA'

    def test_storage_non_ascii_filenames(self):
        filename = u'些公.pdf'
        storage = S3Storage(*self.cred, bucket=self.bucket, storage_class='STANDARD_IA')
        new_file_id = storage.create(
            FILE_CONTENT,
            filename=filename,
            content_type='application/pdf'
        )

        assert new_file_id is not None



from depot.io.utils import FileIntent, _is_fieldstorage_like
from depot.io.interfaces import FileStorage

__all__ = ('TW2FileIntentValidator', )

try:
    import tw2.core
except ImportError:
    pass
else:
    class TW2FileIntentValidator(tw2.core.Validator):
        """Validator for ToscaWidgets2

        Converts the uploaded file to a DEPOT FileIntent
        so that it preserves file details like filename
        and content type when uploaded. This is mostly
        required when working with Sprox that converts
        the cgi.FieldStorage or multipart to a standard BLOB when
        handling uploaded files, causing a loss of all
        file metadata.
        """
        extension = None
        msgs = {
            'required': ('file_required', 'Select a file'),
            'badext': "File name must have '$extension' extension",
        }

        def _convert_to_python(self, value, state=None):
            if is_fieldstorage_like(value):
                if self.required and not getattr(value, 'filename', None):
                    raise tw2.core.ValidationError('required', self)

                if (self.extension is not None
                        and not value.filename.endswith(self.extension)):
                    raise tw2.core.ValidationError('badext', self)
            elif value:
                raise tw2.core.ValidationError('corrupt', self)
            elif self.required:
                raise tw2.core.ValidationError('required', self)

            return FileIntent(*FileStorage.fileinfo(value))



import shutil
import os
import unittest
from depot.io.local import LocalFileStorage

FILE_CONTENT = b'HELLO WORLD'


class TestLocalFileStorage(unittest.TestCase):
    def setUp(self):
        self.fs = LocalFileStorage('./lfs')

    def tearDown(self):
        shutil.rmtree('./lfs', ignore_errors=True)

    def test_creation(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        with open(os.path.join('./lfs', file_id, 'file'), 'rb') as f:
            assert FILE_CONTENT == f.read()

    def test_corrupted_metadata(self):
        file_id = self.fs.create(FILE_CONTENT, 'file.txt')
        f = self.fs.get(file_id)

        with open(f._metadata_path, 'w') as mdf:
            mdf.write('NOT JSON')

        did_detect_corrupted_metadata = False
        try:
            f = self.fs.get(file_id)
        except ValueError:
            did_detect_corrupted_metadata = True

        assert did_detect_corrupted_metadata




from __future__ import absolute_import
import itertools

from sqlalchemy import types
from sqlalchemy import event
from sqlalchemy import orm
from sqlalchemy import inspect
from sqlalchemy.orm import ColumnProperty
from sqlalchemy.orm import Session
from sqlalchemy.orm.attributes import get_history

from depot.manager import DepotManager
from .upload import UploadedFile


class UploadedFileField(types.TypeDecorator):
    """Provides support for storing attachments to **SQLAlchemy** models.

    ``UploadedFileField`` can be used as a Column type to store files
    into the model. The actual file itself will be uploaded to the
    default Storage, and only the :class:`depot.fields.upload.UploadedFile`
    information will be stored on the database.

    The ``UploadedFileField`` is transaction aware, so it will delete
    every uploaded file whenever the transaction is rolled back and will
    delete any old file whenever the transaction is committed. This is
    the reason you should never associate the same :class:`depot.fields.upload.UploadedFile`
    to two different ``UploadedFileField``, otherwise you might delete a file
    already used by another model. It is usually best to just set the ``file``
    or ``bytes`` as content of the column and let the ``UploadedFileField``
    create the :class:`depot.fields.upload.UploadedFile` by itself whenever it's content is set.

    """
    impl = types.Unicode
    cache_ok = False

    def __init__(self, filters=tuple(), upload_type=UploadedFile, upload_storage=None, *args, **kw):
        super(UploadedFileField, self).__init__(*args, **kw)
        self._filters = filters
        self._upload_type = upload_type
        self._upload_storage = upload_storage

    def load_dialect_impl(self, dialect):
        return dialect.type_descriptor(types.VARCHAR(4000))

    def process_bind_param(self, value, dialect):
        if not value:
            return None

        if not isinstance(value, self._upload_type):
            raise ValueError('AttachmentField requires %s, '
                             'got %s instead' % (self._upload_type, type(value)))

        return value.encode()

    def process_result_value(self, value, dialect):
        if not value:
            return None
        return self._upload_type.decode(value)


class _SQLAMutationTracker(object):
    mapped_entities = {}

    @classmethod
    def _field_set(cls, target, value, oldvalue, initiator):
        if value is None or isinstance(value, UploadedFile):
            return value

        inspection = inspect(target)
        set_property = inspection.mapper.get_property(initiator.key)
        column_type = set_property.columns[0].type
        assert(isinstance(column_type, UploadedFileField))

        upload_type = column_type._upload_type
        value = upload_type(value, column_type._upload_storage)
        value._apply_filters(column_type._filters)

        # Mark newly added files on assignment, so that in case of a rollback
        # without going through flush we have track of the files
        # that need to be rolled back.
        session = inspection.session
        if session is not None:
            session._depot_new = getattr(session, '_depot_new', set())
            session._depot_new.update(value.files)
        else:
            # If object is not attached to a session preserve the files
            # until it gets added to one.
            target._depot_new = getattr(target, '_depot_new', set())
            target._depot_new.update(value.files)

        return value

    @classmethod
    def _mapper_configured(cls, mapper, class_):
        for mapper_property in mapper.iterate_properties:
            if isinstance(mapper_property, ColumnProperty):
                for idx, col in enumerate(mapper_property.columns):
                    if isinstance(col.type, UploadedFileField):
                        if idx > 0:
                            # Not clear when this might happen, but ColumnProperty can have
                            # multiple columns assigned. We should probably take the first one
                            # as is done by ColumnProperty.expression but there is no guarantee
                            # that it would be the right thing to do.
                            raise TypeError('UploadedFileField currently supports a single column')
                        cls.mapped_entities.setdefault(class_, []).append(mapper_property.key)
                        event.listen(mapper_property, 'set', cls._field_set, retval=True, propagate=True)

    @classmethod
    def _session_rollback(cls, session, previous_transaction):
        if hasattr(session, '_depot_old'):
            del session._depot_old
        if hasattr(session, '_depot_new'):
            for entry in session._depot_new:
                depot, fileid = entry.split('/', 1)
                depot = DepotManager.get(depot)
                depot.delete(fileid)
            del session._depot_new

    @classmethod
    def _session_committed(cls, session):
        if hasattr(session, '_depot_old'):
            for entry in session._depot_old:
                depot, fileid = entry.split('/', 1)
                depot = DepotManager.get(depot)
                depot.delete(fileid)
            del session._depot_old
        if hasattr(session, '_depot_new'):
            del session._depot_new

    @classmethod
    def _session_flush(cls, session, flush_context, instances):
        for obj in session.deleted:
            class_ = obj.__class__
            tracked_columns = cls.mapped_entities.get(class_, tuple())
            for col in tracked_columns:
                value = getattr(obj, col)
                if value is not None:
                    session._depot_old = getattr(session, '_depot_old', set())
                    session._depot_old.update(value.files)

        for obj in session.new.union(session.dirty):
            class_ = obj.__class__
            tracked_columns = cls.mapped_entities.get(class_, tuple())
            for col in tracked_columns:
                history = get_history(obj, col)
                added_files = itertools.chain(*(f.files for f in history.added
                                               if f is not None))
                deleted_files = itertools.chain(*(f.files for f in history.deleted
                                                 if f is not None))

                session._depot_new = getattr(session, '_depot_new', set())
                session._depot_new.update(added_files)
                session._depot_old = getattr(session, '_depot_old', set())
                session._depot_old.update(deleted_files)

    @classmethod
    def _session_deleted_obj(cls, session, obj):
        # Tracking indirectly deleted objects
        # a relationship.remove, or a query().filter().delete()
        class_ = obj.__class__
        tracked_columns = cls.mapped_entities.get(class_, tuple())
        for col in tracked_columns:
            value = getattr(obj, col)
            if value is not None:
                session._depot_old = getattr(session, '_depot_old', set())
                session._depot_old.update(value.files)

    @classmethod
    def _session_attached(cls, session, instance):
        session._depot_new = getattr(session, '_depot_new', set())
        session._depot_new.update(getattr(instance, '_depot_new', set()))

    @classmethod
    def setup(cls):
        event.listen(orm.Mapper, 'mapper_configured', cls._mapper_configured)
        event.listen(Session, 'after_soft_rollback', cls._session_rollback)
        event.listen(Session, 'after_commit', cls._session_committed)
        event.listen(Session, 'before_attach', cls._session_attached)
        event.listen(Session, 'before_flush', cls._session_flush)
        event.listen(Session, 'persistent_to_deleted', cls._session_deleted_obj)

_SQLAMutationTracker.setup()

try:  # pragma: no cover
    from sprox.sa.widgetselector import SAWidgetSelector
    from tw2.forms import FileField as TW2FileField
    SAWidgetSelector.default_widgets.setdefault(UploadedFileField, TW2FileField)

    from sprox.sa.validatorselector import SAValidatorSelector
    from ..validators import TW2FileIntentValidator
    SAValidatorSelector.default_validators.setdefault(UploadedFileField, TW2FileIntentValidator)
except ImportError:  # pragma: no cover
    pass



"""General interface provided by all storage backends implemented in DEPOT.

Each storage backend is required to respect those classes and implement
the abstractmethods.

"""
from .._compat import with_metaclass
from abc import ABCMeta, abstractmethod, abstractproperty
from io import IOBase
from depot.io.utils import FileIntent, _FileInfo


class StoredFile(IOBase):
    """Interface for already saved files.

    It provides metadata on the stored file through following properties:
        - file_id
        - filename
        - content_type
        - last_modified
        - content_length

    Already stored files can only be read back, so they are required to only provide
    ``read(self, n=-1)``, ``close()`` methods and ``closed`` property so that they
    can be read.

    To replace/overwrite a file content do not try to call the ``write`` method,
    instead use the storage backend to replace the file content.
    """
    def __init__(self, file_id, filename=None, content_type=None, last_modified=None,
                 content_length=None):
        self.file_id = file_id
        self.filename = filename
        self.content_type = content_type
        self.last_modified = last_modified
        self.content_length = content_length

    def readable(self):
        """Returns if the stored file is readable or not

        Usually all stored files are readable
        """
        return True

    def writable(self):
        """Returns if the stored file is writable or not

        Stored files are not writable, you should rely on the relative
        :class:`FileStorage` to overwrite their content
        """
        return False

    def seekable(self):
        """Returns if the stored file is seekable or not

        By default stored files are not seekable
        """
        return False

    @property
    def name(self):
        """This is the filename of the saved file

        If a filename was not available when the file was created
        this will return "unnamed" as filename.
        """
        return self.filename

    @abstractmethod
    def read(self, n=-1):  # pragma: no cover
        """Reads ``n`` bytes from the file.

        If ``n`` is not specified or is ``-1`` the whole
        file content is read in memory and returned
        """
        return

    @abstractmethod
    def close(self, *args, **kwargs):  # pragma: no cover
        """Closes the file.

        After closing the file it won't be possible to read
        from it anymore. Some implementation might not do anything
        when closing the file, but they still are required to prevent
        further reads from a closed file.
        """
        return

    @abstractproperty
    def closed(self):  # pragma: no cover
        """Returns if the file has been closed.

        When ``closed`` return ``True`` it won't be possible
        to read anoymore from this file.
        """
        return

    @property
    def public_url(self):
        """The public HTTP url from which file can be accessed.

        When supported by the storage this will provide the
        public url to which the file content can be accessed.
        In case this returns ``None`` it means that the file can
        only be served by the :class:`DepotMiddleware` itself.
        """
        return None

    def __repr__(self):
        return '<%s:%s filename=%s content_type=%s last_modified=%s>' % (self.__class__.__name__,
                                                                         self.file_id,
                                                                         self.filename,
                                                                         self.content_type,
                                                                         self.last_modified)


class FileStorage(with_metaclass(ABCMeta, object)):
    """Interface for storage providers.

    The FileStorage base class declares a standard interface for storing and retrieving files
    in an underlying storage system.

    Each storage system implementation is required to provide this interface to correctly work
    with filedepot.
    """
    @staticmethod
    def fileid(file_or_id):
        """Gets the ID of a given :class:`StoredFile`

        If the given parameter is already a StoredFile id it will
        directly return it.
        """
        return getattr(file_or_id, 'file_id', file_or_id)

    @staticmethod
    def fileinfo(fileobj, filename=None, content_type=None, existing=None):
        """Tries to extract from the given input the actual file object, filename and content_type

        This is used by the create and replace methods to correctly deduce their parameters
        from the available information when possible.
        """
        return _FileInfo(fileobj, filename, content_type).get_info(existing)

    @abstractmethod
    def get(self, file_or_id):  # pragma: no cover
        """Opens the file given by its unique id.

        This operation is guaranteed to return
        a :class:`StoredFile` instance or should raise ``IOError``
        if the file is not found.
        """
        return

    @abstractmethod
    def create(self, content, filename=None, content_type=None):  # pragma: no cover
        """Saves a new file and returns the ID of the newly created file.

        ``content`` parameter can either be ``bytes``, another ``file object``,
        ``multipart.MultipartPart`` or a :class:`cgi.FieldStorage`.
        When ``filename`` and ``content_type`` parameters are not provided
        they are deducted from the content itself.
        """
        return

    @abstractmethod
    def replace(self, file_or_id, content, filename=None, content_type=None):  # pragma: no cover
        """Replaces an existing file, an ``IOError`` is raised if the file didn't already exist.

        Given a :class:`StoredFile` or its ID it will replace the current content
        with the provided ``content`` value. If ``filename`` and ``content_type`` are
        provided or can be deducted by the ``content`` itself they will also replace
        the previous values, otherwise the current values are kept.
        """
        return

    @abstractmethod
    def delete(self, file_or_id):  # pragma: no cover
        """Deletes a file. If the file didn't exist it will just do nothing."""
        return

    @abstractmethod
    def exists(self, file_or_id):  # pragma: no cover
        """Returns if a file or its ID still exist."""
        return

    @abstractmethod
    def list(self):
        """Returns a list of file IDs that exist in the Storage.

        Depending on the implementation there is the possibility that this returns more IDs
        than there have been created. Therefore this method is NOT guaranteed to be RELIABLE."""
        return []



from typing import List
import uuid
from datetime import datetime
from google.cloud import storage
from depot.io import utils
from depot.io.interfaces import FileStorage, StoredFile
from depot._compat import unicode_text, percent_encode
from google.cloud.exceptions import NotFound
from google.oauth2 import service_account
import os
from depot.utils import make_content_disposition
CANNED_ACL_PUBLIC_READ = 'public-read'
CANNED_ACL_PRIVATE = 'private'

class GCSStoredFile(StoredFile):
    def __init__(self, file_id, blob):
        _check_file_id(file_id)

        self.blob = blob
        self._fileid = file_id
        self._closed = False

        metadata = blob.metadata or {}
        filename = metadata.get('x-depot-filename')
        content_type = metadata.get('x-depot-content-type')
        content_length = blob.size

        last_modified = None
        try:
            last_modified = metadata.get('x-depot-modified')
            if last_modified:
                last_modified = datetime.strptime(last_modified, '%Y-%m-%d %H:%M:%S')
        except:
            pass

        super(GCSStoredFile, self).__init__(file_id, filename, content_type, last_modified, content_length)
        self._pos = 0

    def read(self, n=-1):
        if self.closed:
            raise ValueError("I/O operation on closed file")

        if self._pos == 0:
            # If we are starting a new read, we need to get the latest generation of the blob
            self.blob = storage.Blob(self.blob.name, self.blob.bucket)
        
        data = self.blob.download_as_bytes(start=self._pos, end=self._pos + n - 1 if n != -1 else None)
        self._pos += len(data)
        return data

    def close(self, *args, **kwargs):
        self._closed = True

    @property
    def closed(self):
        return self._closed

    @property
    def public_url(self):
        return self.blob.public_url


class GCSStorage(FileStorage):
    def __init__(self, project_id=None, credentials=None, bucket=None, policy=None, storage_class=None, prefix=''):
        if not credentials:
            if not os.environ.get("GOOGLE_APPLICATION_CREDENTIALS"):
                raise ValueError("GOOGLE_APPLICATION_CREDENTIALS environment variable not set")
            credentials = os.environ["GOOGLE_APPLICATION_CREDENTIALS"]
        
        if isinstance(credentials, dict):
            credentials = service_account.Credentials.from_service_account_info(credentials)
        elif isinstance(credentials, str):
            credentials = service_account.Credentials.from_service_account_file(credentials)
        else:
            credentials = credentials
        
        if not project_id:
            project_id = credentials.project_id

        policy = policy or CANNED_ACL_PUBLIC_READ
        assert policy in [CANNED_ACL_PUBLIC_READ, CANNED_ACL_PRIVATE], (
            "Key policy must be %s or %s" % (CANNED_ACL_PUBLIC_READ, CANNED_ACL_PRIVATE))
        self.client = storage.Client(project=project_id, credentials=credentials)

        
        # check if bucket exists
        try:
            self.bucket = self.client.get_bucket(bucket)
        except NotFound:
            self.bucket = self.client.create_bucket(bucket)

        self._policy = policy or CANNED_ACL_PUBLIC_READ
        self._storage_class = storage_class or 'STANDARD'
        self._prefix = prefix

        if policy == CANNED_ACL_PUBLIC_READ:
            self.set_bucket_public_iam(self.bucket)

    def get(self, file_or_id):
        file_id = self.fileid(file_or_id)
        _check_file_id(file_id)

        blob = self.bucket.blob(self._prefix+file_id)
        if not blob.exists():
            # TODO: Can we get rid of this extra exists check?
            raise IOError('File %s not existing' % file_id)

        blob.reload()
        return GCSStoredFile(file_id, blob)
    
    def set_bucket_public_iam(self, bucket, members=("allUsers", )):
        policy = bucket.get_iam_policy(requested_policy_version=3)
        policy.bindings.append(
            {"role": "roles/storage.objectViewer", "members": members}
        )
        bucket.set_iam_policy(policy)
        
    def __save_file(self,file_id, content, filename, content_type=None):
        if filename:
            filename = percent_encode(filename, safe='!#$&+-.^_`|~', encoding='utf-8')
        blob = self.bucket.blob(self._prefix+file_id)
        blob.content_type = content_type
        #blob.content_encoding = 'utf-8'
        #blob.cache_control = 'no-cache'
        blob.content_disposition = make_content_disposition('inline', filename)

        blob.metadata = {
            'x-depot-modified': utils.timestamp(),
            'x-depot-filename': filename,
            'x-depot-content-type': content_type
        }

        if hasattr(content, 'read'):
            try:
                pos = content.tell()
                content.seek(pos)
                blob.upload_from_file(content, content_type=content_type)
            except:
                content = content.read()
                blob.upload_from_string(content, content_type=content_type)
        else:
            if isinstance(content, unicode_text):
                raise TypeError('Only bytes can be stored, not unicode')
            blob.upload_from_string(content, content_type=content_type)
        
        if self._policy == CANNED_ACL_PUBLIC_READ:
            blob.make_public()

    def create(self, content, filename=None, content_type=None):
        content, filename, content_type = self.fileinfo(content, filename, content_type)
        new_file_id = str(uuid.uuid4())
        self.__save_file(new_file_id, content, filename, content_type)
        return new_file_id

    def replace(self, file_or_id, content, filename=None, content_type=None):
        file_id = self.fileid(file_or_id)
        _check_file_id(file_id)
        
        if isinstance(file_or_id, StoredFile) and file_or_id is content:
            # This is a backup, no need to check if file exists.
            pass
        elif not self.exists(file_id):
            # Check file existed and we are not using replace
            # as a way to force a specific file id on creation.
            raise IOError('File %s not existing' % file_id)

        content, filename, content_type = self.fileinfo(content, filename, content_type,
                                                        lambda: self.get(file_id))
        self.__save_file(file_id, content, filename, content_type)

    def delete(self, file_or_id):
        file_id = self.fileid(file_or_id)
        _check_file_id(file_id)

        blob = self.bucket.blob(self._prefix+file_id)
        generation_match_precondition = None
        try:
            blob.reload()
            generation_match_precondition = blob.generation
            blob.delete(if_generation_match=generation_match_precondition)
        except NotFound:
            pass

    def exists(self, file_or_id):
        file_id = self.fileid(file_or_id)
        _check_file_id(file_id)

        blob = self.bucket.blob(self._prefix+file_id)
        return blob.exists()

    def list(self):
        return [blob.name for blob in self.bucket.list_blobs()]


def _check_file_id(file_id):
    # Check that the given file id is valid, this also
    # prevents unsafe paths.
    try:
        uuid.UUID('{%s}' % file_id)
    except:
        raise ValueError('Invalid file id %s' % file_id)


import mimetypes
import os
from datetime import datetime
from tempfile import SpooledTemporaryFile
from depot._compat import byte_string


INMEMORY_FILESIZE = 1024*1024


def timestamp():
    return datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')


def file_from_content(content):
    """Provides a real file object from file content

    Converts ``FileStorage``, ``FileIntent`` and
    ``bytes`` to an actual file.
    """
    must_close = False
    f = content
    if isinstance(content, FileIntent):
        f = content._fileobj
    elif isinstance(content, byte_string):
        must_close = True
        f = SpooledTemporaryFile(INMEMORY_FILESIZE)
        f.write(content)
    elif _is_fieldstorage_like(content):
        f = content.file
    f.seek(0)
    return must_close, f


class FileIntent(object):
    """Represents the intention to upload a file

    Whenever a file can be stored by depot, a FileIntent
    can be passed instead of the file itself. This permits
    to easily upload objects that are not files or to add
    missing information to the uploaded files.
    """
    def __init__(self, fileobj, filename, content_type):
        self._fileobj = fileobj
        self._filename = filename
        self._content_type = content_type

    @property
    def fileinfo(self):
        return self._fileobj, self._filename, self._content_type


class _FileInfo(object):
    """Utility class to get a file object, filename and content_type from available data.

    This is used by most Storage to convert data to a file like object, to
    find a meaningful filename for the file and to detect the content type of the
    file.
    """
    DEFAULT_CONTENT_TYPE = 'application/octet-stream'
    DEFAULT_NAME = 'unnamed'

    def __init__(self, fileobj, filename=None, content_type=None):
        self._content, self._filename, self._content_type = self._resolve(fileobj,
                                                                          filename,
                                                                          content_type)

    def get_info(self, existing=None):
        if self._filename is None and existing is not None:
            if callable(existing):
                existing = existing()
            return self._content, existing.filename, existing.content_type

        return (
            self._content,
            self._filename or self.DEFAULT_NAME,
            self._content_type or self.DEFAULT_CONTENT_TYPE
        )

    def _resolve(self, fileobj, filename, content_type):
        if isinstance(fileobj, FileIntent):
            return fileobj.fileinfo

        content = self._get_content_from_file_obj(fileobj)
        filename = filename or self._get_filename_from_fileob(fileobj)
        content_type = content_type or self._get_content_type_from_fileobj(fileobj)

        if content_type is None and filename is not None:
            content_type = mimetypes.guess_type(filename, strict=False)[0]

        return content, filename, content_type

    def _get_content_from_file_obj(self, fileobj):
        if _is_fieldstorage_like(fileobj):
            return fileobj.file
        return fileobj

    def _get_filename_from_fileob(self, fileobj):
        if getattr(fileobj, 'filename', None) is not None:
            return fileobj.filename
        elif getattr(fileobj, 'name', None) is not None:
            return os.path.basename(fileobj.name)

    def _get_content_type_from_fileobj(self, fileobj):
        if getattr(fileobj, 'content_type', None) is not None:
            return fileobj.content_type
        elif getattr(fileobj, 'type', None) is not None:
            return fileobj.type


def _is_fieldstorage_like(obj):
    # Detect cgi.FieldStorage and multipart modules
    return (getattr(obj, 'filename', None) is not None and \
            getattr(obj, 'file', None) not in (None, False))



# -*- coding: utf-8 -*-
import shutil
import unittest
import tempfile, os, cgi, base64
from PIL import Image
from depot.fields.filters.thumbnails import WithThumbnailFilter
from depot.fields.ming import UploadedFileProperty
from depot.manager import DepotManager, get_file
from .base_ming import setup_database, clear_database, DBSession
from ming.odm.declarative import MappedClass
from ming.odm import FieldProperty
from ming import schema as s, Field
from .utils import create_cgifs, OpenFiles
from depot.fields.specialized.image import UploadedImageWithThumb
from depot._compat import u_

of = OpenFiles()


def setUpModule():
    setup_database()

    DepotManager._clear()
    DepotManager.configure('default', {'depot.storage_path': './lfs'})
    DepotManager.configure('another', {'depot.storage_path': './lfs'})
    DepotManager.alias('another_alias', 'another')
    DepotManager.make_middleware(None)


def tearDownModule():
    shutil.rmtree('./lfs', ignore_errors=True)


class Document(MappedClass):
    class __mongometa__:
        session = DBSession
        name = 'depot_test_document'

    _id = FieldProperty(s.ObjectId)
    name = FieldProperty(str)
    content = UploadedFileProperty()
    photo = UploadedFileProperty(upload_type=UploadedImageWithThumb)
    second_photo = UploadedFileProperty(filters=(WithThumbnailFilter((12, 12), 'PNG'),))
    targeted_content = UploadedFileProperty(upload_storage='another_alias')


class TestMingAttachments(unittest.TestCase):
    def setUp(self):
        self.file_content = b'this is the file content'
        self.fake_file = of.add(tempfile.NamedTemporaryFile())
        self.fake_file.write(self.file_content)
        self.fake_file.flush()
        self.fake_file.seek(0)
        clear_database()

    def tearDown(self):
        of.close_all()

    def test_accessing_class_property(self):
        # This is to check for regression in a bug in property descriptor
        prop = Document.content
        assert isinstance(prop, FieldProperty), prop

    def test_create_fromfile(self):
        doc = Document(name='Foo', content = of.open(self.fake_file.name, 'rb'))
        DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()
        assert d.content.file.read() == self.file_content
        assert d.content.file.filename == os.path.basename(self.fake_file.name)

    def test_create_fromfile_flush_single_document(self):
        doc = Document(name='Foo', content = of.open(self.fake_file.name, 'rb'))
        DBSession.flush(doc)
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()
        assert d.content.file.read() == self.file_content
        assert d.content.file.filename == os.path.basename(self.fake_file.name)

    def test_edit_existing(self):
        doc = Document(name=u_('Foo2'))
        with open(self.fake_file.name, 'rb') as f:
            doc.content = f
            DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo2')).first()
        old_file = d.content.path

        d.content = b'HELLO'
        new_file = d.content.path

        DBSession.flush()
        DBSession.clear()

        assert get_file(new_file).read() == b'HELLO'

        try:
            fold = get_file(old_file)
            assert False, 'Should have raised IOError here'
        except IOError:
            pass

    def test_edit_existing_rollback(self):
        doc = Document(name=u_('Foo3'))
        with open(self.fake_file.name, 'rb') as f:
            doc.content = f
            DBSession.flush()
        DBSession.clear()


        d = Document.query.find(dict(name='Foo3')).first()
        old_file = d.content.path

        d.content = b'HELLO'
        new_file = d.content.path

        DBSession.clear()
        assert get_file(old_file).read() == self.file_content

        self.skipTest("Currently Ming Doesn't provide a way to handle discarded documents")
        try:
            fold = get_file(new_file)
            assert False, 'Should have raised IOError here'
        except IOError:
            pass

    def test_create_fromfield(self):
        field = create_cgifs('image/jpeg', self.fake_file, 'test.jpg')
        of.add(field.file)

        doc = Document(name=u_('Foo'), content=field)
        DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()
        assert d.content.file.read() == self.file_content
        assert d.content.filename == 'test.jpg'
        assert d.content.content_type == 'image/jpeg', d.content.content_type
        assert d.content.url == '/depot/%s' % d.content.path

    def test_create_empty(self):
        doc = Document(name=u_('Foo'), content=None)
        DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()
        assert d.content is None

    def test_delete_existing(self):
        doc = Document(name=u_('Foo2'))
        with open(self.fake_file.name, 'rb') as f:
            doc.content = f
            DBSession.flush()
        DBSession.clear()


        d = Document.query.find(dict(name='Foo2')).first()
        old_file = d.content.path
        d.query.delete()

        DBSession.flush()
        DBSession.clear()

        with self.assertRaises(IOError):
            fold = get_file(old_file)

    def test_delete_existing_rollback(self):
        doc = Document(name=u_('Foo3'))
        with open(self.fake_file.name, 'rb') as f:
            doc.content = f
            DBSession.flush()
        DBSession.clear()


        d = Document.query.find(dict(name='Foo3')).first()
        old_file = d.content.path
        d.delete()
        DBSession.clear()

        assert get_file(old_file).read() == self.file_content

    def test_create_with_alias(self):
        with open(self.fake_file.name, 'rb') as f:
            doc = Document(name='Foo', targeted_content=f)
            DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()
        assert d.targeted_content.file.read() == self.file_content
        assert d.targeted_content.file.filename == os.path.basename(self.fake_file.name)
        assert d.targeted_content.depot_name == 'another'


class TestMingImageAttachments(unittest.TestCase):
    def setUp(self):
        self.file_content = b'''R0lGODlhEQAUAPcAAC4uLjAwMDIyMjMzMjQ0NDU1NDY2Njk2Mzg4ODo6Oj49Ozw8PD4+PkE+OEA/PkhAN0tCNk5JPFFGNV1KMFhNNVFHOFJJPVVLPVhOPXZfKXVcK2ZQNGZXNGtZMnNcNHZeNnldMHJfOn1hKXVjOH9oO0BAQEJCQkREREVFREZGRklGQ05KQ0hISEpKSkxMTE5OTlZRSlFQT19XSFBQUFJSUlRUVGFUQmFVQ2ZZQGtdQnNiQqJ/HI1uIYBnLIllKoZrK4FqLoVqL4luLIpsLpt7J515JJ50KZhzLYFnMIFlM4ZlMIFkNI1uNoJoOoVrPIlvO49yMolwPpB2O5p4Op98PaB3IKN4JqN8J6h7I6J5LaZ+LLF+ILGGG7+JG72OGLKEI7aHIrOEJL2JI7mMN76YNcGJG8SOG8WLHMONH86eEs+aFsGSG8eQHMySG9uVFduXFdeeE9eaFdScF96YE9yaFOKcEuOdEtWgFNiiEduhE96pEuqlD+qmD+KpD+yoDu6rDuysDvCuDfCvDeuwD/SzC/a2CvGwDfKxDPi5Cfi5Cvq8CeCjEuehEOagEeijEOKoEOStFMOOK8+TLM6YNNChItGgLtylKt6gMNqgON6jPOChLfi/JOSrNeGvN9KhRtykRNWkSOCnQOCpSOawQue1T+a6Su67SOGsUO/AVAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAAAAAAALAAAAAARABQAAAj+AAEIHEiwoMAACBMqXIhQgMOHDwdAfEigYsUCT6KQScNjxwGLBAyINPBhCilUmxIV6uMlw0gEMJeMGWWqFCU8hA4JEgETQYIEDcRw6qRlwgMIGvJwkfAzwYIFXQBB8qHg6VMKFawuYNBBjaIqDhhI+cGgrNmyJXooGrShBJBKcIpwiFCibl0TehDdsWBCiBxLZuIwolMGiwcTJ4gUenThBAokSVRgGFJnzhYQJ1KsyRkkhWfPK87McWPEM4sRhgItCsGitQ5PmtxYUdK6BY4rf/zwYRNmUihRpyQdaUHchQsoX/Y4amTnUqZPoG7YMO7ihfUcYNC0eRMJExMqMKweW59BfkYMEk2ykHAio3x5GvDjy58Pv4b9+/jz2w8IADs='''
        self.file_content = base64.b64decode(self.file_content)
        self.fake_file = of.add(tempfile.NamedTemporaryFile())
        self.fake_file.write(self.file_content)
        self.fake_file.flush()
        self.fake_file.seek(0)
        clear_database()

    def tearDown(self):
        of.close_all()

    def test_create_fromfile(self):
        doc = Document(name=u_('Foo'))
        with open(self.fake_file.name, 'rb') as f:
            doc.photo = f
            DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()
        assert d.photo.file.read() == self.file_content
        assert d.photo.filename == os.path.basename(self.fake_file.name)

    def test_create_fromfield(self):
        field = cgi.FieldStorage()
        field.filename = u_('àèìòù.gif')
        with open(self.fake_file.name, 'rb') as f:
            field.file = f
            doc = Document(name=u_('Foo'), photo=field)
            DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()
        assert d.photo.file.read() == self.file_content
        assert d.photo.filename == field.filename
        assert d.photo.url == '/depot/%s' % d.photo.path
        assert d.photo.thumb_url == '/depot/%s' % d.photo.thumb_path
        assert d.photo.url != d.photo.thumb_url

    def test_thumbnail(self):
        field = create_cgifs('image/gif', self.fake_file, 'test.gif')
        of.add(field.file)

        doc = Document(name=u_('Foo'), photo=field)
        DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()
        assert os.path.exists(d.photo.thumb_file._file_path)

        thumb = Image.open(d.photo.thumb_file._file_path)
        thumb.verify()
        assert thumb.format.upper() == d.photo.thumbnail_format.upper()

    def test_public_url(self):
        doc = Document(name=u_('Foo'))
        doc.photo = of.open(self.fake_file.name, 'rb')
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()

        # This is to edit the saved data bypassing UploadedFileProperty
        photo = FieldProperty(Field('photo', s.Anything)).__get__(d, Document)
        photo['_public_url'] = 'PUBLIC_URL'
        photo['_thumb_public_url'] = 'THUMB_PUBLIC_URL'

        # Now check that depot does the right thing when public urls are available
        assert d.photo.url == 'PUBLIC_URL'
        assert d.photo.thumb_url == 'THUMB_PUBLIC_URL'

    def test_rollback(self):
        self.skipTest("Currently Ming Doesn't provide a way to handle discarded documents")

        doc = Document(name=u_('Foo3'))
        doc.photo = of.open(self.fake_file.name, 'rb')

        uploaded_file = doc.photo.path
        uploaded_thumb = doc.photo.thumb_path

        DBSession.clear()

        try:
            fold = get_file(uploaded_file)
            assert False, 'Should have raised IOError here'
        except IOError:
            pass

        try:
            fold = get_file(uploaded_thumb)
            assert False, 'Should have raised IOError here'
        except IOError:
            pass


class TestMingThumbnailFilter(unittest.TestCase):
    def setUp(self):
        self.file_content = b'''R0lGODlhEQAUAPcAAC4uLjAwMDIyMjMzMjQ0NDU1NDY2Njk2Mzg4ODo6Oj49Ozw8PD4+PkE+OEA/PkhAN0tCNk5JPFFGNV1KMFhNNVFHOFJJPVVLPVhOPXZfKXVcK2ZQNGZXNGtZMnNcNHZeNnldMHJfOn1hKXVjOH9oO0BAQEJCQkREREVFREZGRklGQ05KQ0hISEpKSkxMTE5OTlZRSlFQT19XSFBQUFJSUlRUVGFUQmFVQ2ZZQGtdQnNiQqJ/HI1uIYBnLIllKoZrK4FqLoVqL4luLIpsLpt7J515JJ50KZhzLYFnMIFlM4ZlMIFkNI1uNoJoOoVrPIlvO49yMolwPpB2O5p4Op98PaB3IKN4JqN8J6h7I6J5LaZ+LLF+ILGGG7+JG72OGLKEI7aHIrOEJL2JI7mMN76YNcGJG8SOG8WLHMONH86eEs+aFsGSG8eQHMySG9uVFduXFdeeE9eaFdScF96YE9yaFOKcEuOdEtWgFNiiEduhE96pEuqlD+qmD+KpD+yoDu6rDuysDvCuDfCvDeuwD/SzC/a2CvGwDfKxDPi5Cfi5Cvq8CeCjEuehEOagEeijEOKoEOStFMOOK8+TLM6YNNChItGgLtylKt6gMNqgON6jPOChLfi/JOSrNeGvN9KhRtykRNWkSOCnQOCpSOawQue1T+a6Su67SOGsUO/AVAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAAAAAAALAAAAAARABQAAAj+AAEIHEiwoMAACBMqXIhQgMOHDwdAfEigYsUCT6KQScNjxwGLBAyINPBhCilUmxIV6uMlw0gEMJeMGWWqFCU8hA4JEgETQYIEDcRw6qRlwgMIGvJwkfAzwYIFXQBB8qHg6VMKFawuYNBBjaIqDhhI+cGgrNmyJXooGrShBJBKcIpwiFCibl0TehDdsWBCiBxLZuIwolMGiwcTJ4gUenThBAokSVRgGFJnzhYQJ1KsyRkkhWfPK87McWPEM4sRhgItCsGitQ5PmtxYUdK6BY4rf/zwYRNmUihRpyQdaUHchQsoX/Y4amTnUqZPoG7YMO7ihfUcYNC0eRMJExMqMKweW59BfkYMEk2ykHAio3x5GvDjy58Pv4b9+/jz2w8IADs='''
        self.file_content = base64.b64decode(self.file_content)
        self.fake_file = of.add(tempfile.NamedTemporaryFile())
        self.fake_file.write(self.file_content)
        self.fake_file.flush()
        self.fake_file.seek(0)
        clear_database()

    def tearDown(self):
        of.close_all()

    def test_create_fromfile(self):
        doc = Document(name=u_('Foo'))
        doc.second_photo = of.open(self.fake_file.name, 'rb')
        DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()
        assert d.second_photo.file.read() == self.file_content
        assert d.second_photo.filename == os.path.basename(self.fake_file.name)

    def test_create_fromfield(self):
        field = cgi.FieldStorage()
        field.filename = u_('àèìòù.gif')
        field.file = of.open(self.fake_file.name, 'rb')

        doc = Document(name=u_('Foo'), second_photo=field)
        DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()
        assert d.second_photo.file.read() == self.file_content
        assert d.second_photo.filename == field.filename
        assert d.second_photo.url == '/depot/%s' % d.second_photo.path
        assert d.second_photo.thumb_12x12_url == '/depot/%s' % d.second_photo.thumb_12x12_path
        assert d.second_photo.url != d.second_photo.thumb_12x12_url

    def test_thumbnail(self):
        field = create_cgifs('image/gif', self.fake_file, 'test.gif')
        of.add(field.file)

        doc = Document(name=u_('Foo'), second_photo=field)
        DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()
        thumbnail_local_path = DepotManager.get_file(d.second_photo.thumb_12x12_path)._file_path

        assert os.path.exists(thumbnail_local_path)

        thumb = Image.open(thumbnail_local_path)
        thumb.verify()
        assert thumb.format.upper() == 'PNG'
        assert max(thumb.size) == 12

    def test_public_url(self):
        doc = Document(name=u_('Foo'))
        doc.second_photo = of.open(self.fake_file.name, 'rb')
        doc.content = of.open(self.fake_file.name, 'rb')
        DBSession.flush()
        DBSession.clear()

        d = Document.query.find(dict(name='Foo')).first()

        # This is to edit the saved data bypassing UploadedFileProperty
        second_photo = FieldProperty(Field('second_photo', s.Anything)).__get__(d, Document)
        second_photo['_public_url'] = 'PUBLIC_URL'

        # Now check that depot does the right thing when public urls are available
        assert d.second_photo.url == 'PUBLIC_URL'
        assert d.second_photo.thumb_12x12_url.startswith('/depot/default/')

    def test_rollback(self):
        self.skipTest("Currently Ming Doesn't provide a way to handle discarded documents")

        doc = Document(name=u_('Foo3'))
        doc.second_photo = of.open(self.fake_file.name, 'rb')

        uploaded_file = doc.second_photo.path
        uploaded_thumb = doc.second_photo.thumb_12x12_path

        DBSession.clear()

        try:
            fold = get_file(uploaded_file)
            assert False, 'Should have raised IOError here'
        except IOError:
            pass

        try:
            fold = get_file(uploaded_thumb)
            assert False, 'Should have raised IOError here'
        except IOError:
            pass



from depot.io import utils
from depot.manager import DepotManager
from ..interfaces import FileFilter
from PIL import Image
from io import BytesIO


class WithThumbnailFilter(FileFilter):
    """Uploads a thumbnail together with the file.

    Takes for granted that the file is an image.
    The resulting uploaded file will provide three additional
    properties named:

        - ``thumb_X_id`` -> The depot file id
        - ``thumb_X_path`` -> Where the file is available in depot
        - ``thumb_X_url`` -> Where the file is served.

    Where ``X`` is the resolution specified as ``size`` in the
    filter initialization. By default this is ``(128, 128)``?so
    you will get ``thumb_128x128_id``, ``thumb_128x128_url`` and
    so on.

    .. warning::

        Requires Pillow library

    """
    def __init__(self, size=(128,128), format='PNG'):
        self.thumbnail_size = size
        self.thumbnail_format = format

    def on_save(self, uploaded_file):
        close_content, content = utils.file_from_content(uploaded_file.original_content)

        try:
            thumbnail = Image.open(content)
            thumbnail.thumbnail(self.thumbnail_size, Image.BILINEAR)
            thumbnail = thumbnail.convert('RGBA')
            thumbnail.format = self.thumbnail_format

            output = BytesIO()
            thumbnail.save(output, self.thumbnail_format)
            output.seek(0)

            thumb_name = 'thumb_%sx%s' % self.thumbnail_size
            thumb_file_name = '%s.%s' % (thumb_name, self.thumbnail_format.lower())
            thumb_path, thumb_id = uploaded_file.store_content(output, thumb_file_name)
            uploaded_file[thumb_name + '_id'] = thumb_id
            uploaded_file[thumb_name + '_path'] = thumb_path
            uploaded_file[thumb_name + '_url'] = DepotManager.get_middleware().url_for(thumb_path)
        finally:
            if close_content:
                content.close()


import platform, sys


if platform.system() == 'Windows':  # pragma: no cover
    WIN = True
else:
    WIN = False

# True if we are running on Python 2.
PY2 = sys.version_info[0] == 2


if not PY2:  # pragma: no cover
    from urllib.parse import quote, unquote

    string_type = str
    unicode_text = str
    byte_string = bytes
    wsgi_string = str

    def u_(s):
        return str(s)

    def bytes_(s):
        return str(s).encode('ascii', 'strict')

    def percent_encode(string, safe, encoding):
        return quote(string, safe, encoding, errors='strict')

    def percent_decode(string):
        return unquote(string)

else:  # pragma: no cover
    from urllib import quote, unquote

    string_type = basestring
    unicode_text = unicode
    byte_string = str
    wsgi_string = str

    def u_(s):
        return unicode(s, 'utf-8')

    def bytes_(s):
        return str(s)

    def percent_encode(string, **kwargs):
        encoding = kwargs.pop('encoding')
        return quote(string.encode(encoding), **kwargs)

    def percent_decode(string):
        return unquote(string)


def with_metaclass(meta, base=object):
    """Create a base class with a metaclass."""
    return meta("NewBase", (base,), {})



from __future__ import absolute_import

from ming.odm import mapper
from ming.odm.property import FieldProperty
from ming.odm.base import session, state, ObjectState
from ming.odm.odmsession import SessionExtension
from ming.schema import Anything

from depot.manager import DepotManager
from .upload import UploadedFile


class _UploadedFileSchema(Anything):
    pass


class UploadedFileProperty(FieldProperty):
    """Provides support for storing attachments to **Ming** MongoDB models.

    ``UploadedFileProperty`` can be used as a field type to store files
    into the model. The actual file itself will be uploaded to the
    default Storage, and only the :class:`depot.fields.upload.UploadedFile`
    information will be stored on the database.

    The ``UploadedFileProperty`` is UnitOfWork aware, so it will delete
    every uploaded file whenever unit of work is flushed and deletes a Document
    that stored files or changes the field of a document storing files. This is
    the reason you should never associate the same :class:`depot.fields.upload.UploadedFile`
    to two different ``UploadedFileProperty``, otherwise you might delete a file
    already used by another document. It is usually best to just set the ``file``
    or ``bytes`` as content of the column and let the ``UploadedFileProperty``
    create the :class:`depot.fields.upload.UploadedFile` by itself whenever it's content is set.

    .. warning::

        As the Ming UnitOfWork does not notify any event in case it gets cleared instead
        of being flushed all the files uploaded before clearing the unit of work will be
        already uploaded but won't have a document referencing them anymore, so DEPOT will
        be unable to delete them for you.

    """
    def __init__(self,  filters=tuple(), upload_type=UploadedFile, upload_storage=None):
        FieldProperty.__init__(self, _UploadedFileSchema())
        self._filters = filters
        self._upload_type = upload_type
        self._upload_storage = upload_storage

    def __set__(self, instance, value):
        if value is not None and not isinstance(value, UploadedFile):
            upload_type = self._upload_type
            value = upload_type(value, self._upload_storage)

        if isinstance(value, UploadedFile):
            value._apply_filters(self._filters)

        old_value = self.__get__(instance, instance.__class__)
        DepotExtension.get_depot_history(instance).swap(old_value, value)
        return FieldProperty.__set__(self, instance, value)

    def __get__(self, instance, owner=None):
        try:
            value = FieldProperty.__get__(self, instance, owner)
        except AttributeError:
            value = None

        if not value:
            return None

        if instance is None:
            return value

        return self._upload_type(value)

    """
    # Looks like this should do nothing on ming.
    def __delete__(self, instance, owner=None):
        old_value = self.__get__(instance, instance.__class__)
        DepotExtension.get_depot_history(instance).delete(old_value)
        return FieldProperty.__delete__(self, instance, owner)
    """


class DepotExtension(SessionExtension):
    """Extends the Ming Session to track files.

    Deletes old files when an entry gets removed or replaced,
    apply this as a Ming ``SessionExtension`` according to Ming
    documentation.
    """
    @classmethod
    def get_depot_history(cls, instance):
        istate = state(instance)
        if not hasattr(istate, '_depot_history'):
            istate._depot_history = _DepotHistory()
        return istate._depot_history

    def _check_object_deleted(self, obj):
        hist = self.get_depot_history(obj)
        if state(obj).status == ObjectState.deleted:
            for prop in mapper(obj).properties:
                if isinstance(prop, UploadedFileProperty):
                    current_value = prop.__get__(obj, obj.__class__)
                    hist.delete(current_value)
            self._flush_object(obj)

    def _flush_object(self, obj):
        history = self.get_depot_history(obj)
        for entry in history.deleted:
            depot, fileid = entry.split('/', 1)
            depot = DepotManager.get(depot)
            depot.delete(fileid)
        history.clear()

    def before_flush(self, obj=None):
        if obj:
            self._check_object_deleted(obj)
        else:
            for class_, id_, obj in self.session.imap:
                self._check_object_deleted(obj)

    def after_flush(self, obj=None):
        if obj:
            self._flush_object(obj)
        else:
            for class_, id_, obj in self.session.imap:
                self._flush_object(obj)


class _DepotHistory(object):
    def __init__(self):
        self.clear()

    def _extract_files(self, obj):
        return obj['files']

    def add(self, obj):
        if obj is None:
            return

        files = self._extract_files(obj)
        self.deleted.difference_update(obj)
        self.new.update(files)

    def delete(self, obj):
        if obj is None:
            return

        files = self._extract_files(obj)
        self.new.difference_update(obj)
        self.deleted.update(files)

    def swap(self, old, new):
        self.delete(old)
        self.add(new)

    def clear(self):
        self.deleted = set()
        self.new = set()


try:  # pragma: no cover
    from sprox.mg.widgetselector import MingWidgetSelector
    from tw2.forms import FileField as TW2FileField
    MingWidgetSelector.default_widgets.setdefault(_UploadedFileSchema, TW2FileField)

    from sprox.mg.validatorselector import MingValidatorSelector
    from ..validators import TW2FileIntentValidator
    MingValidatorSelector.default_validators.setdefault(_UploadedFileSchema, TW2FileIntentValidator)
except ImportError:  # pragma: no cover
    pass


from depot.io import utils
from depot.manager import DepotManager
from ..upload import UploadedFile
from depot.io.interfaces import FileStorage
from PIL import Image
from depot.io.utils import INMEMORY_FILESIZE
from tempfile import SpooledTemporaryFile


class UploadedImageWithThumb(UploadedFile):
    """Uploads an Image with thumbnail.

    The default thumbnail format and size are PNG@128x128, those can be changed
    by inheriting the ``UploadedImageWithThumb`` and replacing the
    ``thumbnail_format`` and ``thumbnail_size`` class properties.

    The Thumbnail file is accessible as ``.thumb_file`` while the
    thumbnail url is ``.thumb_url``.

    .. warning::

        Requires Pillow library

    """

    max_size = 1024
    thumbnail_format = 'PNG'
    thumbnail_size = (128, 128)

    def process_content(self, content, filename=None, content_type=None):
        orig_content = content
        temporary_content = False
        __, filename, content_type = FileStorage.fileinfo(orig_content)
        close_content, content = utils.file_from_content(content)

        try:
            uploaded_image = Image.open(content)
            if max(uploaded_image.size) >= self.max_size:
                uploaded_image.thumbnail((self.max_size, self.max_size), Image.BILINEAR)
                content = SpooledTemporaryFile(INMEMORY_FILESIZE)
                close_content = True
                uploaded_image.save(content, uploaded_image.format)

            content.seek(0)
            super(UploadedImageWithThumb, self).process_content(content, filename, content_type)
        finally:
            if close_content:
                content.close()

        thumbnail = uploaded_image.copy()
        thumbnail.thumbnail(self.thumbnail_size, Image.LANCZOS)
        thumbnail = thumbnail.convert('RGBA')
        thumbnail.format = self.thumbnail_format

        with SpooledTemporaryFile(INMEMORY_FILESIZE) as output:
            thumbnail.save(output, self.thumbnail_format)
            output.seek(0)

            thumb_path, thumb_id = self.store_content(output,
                                                    'thumb.%s' % self.thumbnail_format.lower())
            self['thumb_id'] = thumb_id
            self['thumb_path'] = thumb_path

            thumbnail_file = self.thumb_file
            self['_thumb_public_url'] = thumbnail_file.public_url

    @property
    def thumb_file(self):
        return self.depot.get(self.thumb_id)

    @property
    def thumb_url(self):
        public_url = self['_thumb_public_url']
        if public_url:
            return public_url
        return DepotManager.get_middleware().url_for(self['thumb_path'])



"""
Provides MemoryStorage implementation to keep files in memory.

This is useful for caches or tests.

"""
import uuid
import io
from datetime import datetime

from .interfaces import FileStorage, StoredFile
from . import utils
from .._compat import unicode_text


class MemoryStoredFile(StoredFile):
    def __init__(self, files, file_id, file_data):
        _check_file_id(file_id)
        self._files = files
        self._files_key = file_id
        self._file = None
        super(MemoryStoredFile, self).__init__(file_id=file_id, **file_data['metadata'])

    def read(self, n=-1):
        if self._file is None:
            self._file = io.BytesIO(self._files[self._files_key]['data'])
        return self._file.read(n)

    def close(self):
        if self._file is None:
            self._file = io.BytesIO(self._files[self._files_key]['data'])
        self._file.close()

    @property
    def closed(self):
        if self._file is None:
            return False
        return self._file.closed


class MemoryFileStorage(FileStorage):
    """:class:`depot.io.interfaces.FileStorage` implementation that keeps files in memory.

    This is generally useful for caches and tests.
    """
    def __init__(self, **kwargs):
        self.files = {}

    def get(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        try:
            file_data = self.files[fileid]
        except KeyError:
            raise IOError('File %s not existing' % fileid)

        return MemoryStoredFile(self.files, fileid, file_data)

    def __save_file(self, file_id, content, filename, content_type=None):
        if hasattr(content, 'read'):
            data = content.read()
        else:
            if isinstance(content, unicode_text):
                raise TypeError('Only bytes can be stored, not unicode')
            data = content

        self.files[file_id] = {
            'data': data,
            'metadata': {
                'filename': filename or 'unknown',
                'content_type': content_type,
                'content_length': len(data),
                'last_modified': datetime.strptime(utils.timestamp(), '%Y-%m-%d %H:%M:%S')
            }
        }

    def create(self, content, filename=None, content_type=None):
        new_file_id = str(uuid.uuid1())
        content, filename, content_type = self.fileinfo(content, filename, content_type)
        self.__save_file(new_file_id, content, filename, content_type)
        return new_file_id

    def replace(self, file_or_id, content, filename=None, content_type=None):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        if isinstance(file_or_id, StoredFile) and file_or_id is content:
            # This is a backup, no need to check if file exists.
            pass
        elif not self.exists(fileid):
            # Check file existed and we are not using replace
            # as a way to force a specific file id on creation.
            raise IOError('File %s not existing' % fileid)

        content, filename, content_type = self.fileinfo(content, filename, content_type,
                                                        lambda: self.get(fileid))

        self.delete(fileid)
        self.__save_file(fileid, content, filename, content_type)
        return fileid

    def delete(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)
        self.files.pop(fileid, None)

    def exists(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)
        return fileid in self.files

    def list(self):
        return list(self.files.keys())


def _check_file_id(file_id):
    # Check that the given file id is valid, this also
    # prevents unsafe paths.
    try:
        uuid.UUID('{%s}' % file_id)
    except:
        raise ValueError('Invalid file id %s' % file_id)



# -*- coding: utf-8 -*-
import unittest
from depot.manager import DepotManager


class TestDepotManager(unittest.TestCase):
    def setUp(self):
        DepotManager._clear()

    def test_first_configured_is_default(self):
        DepotManager.configure('first', {'depot.storage_path': './lfs'})
        DepotManager.configure('second', {'depot.storage_path': './lfs2'})
        assert DepotManager.get_default() == 'first'

    def test_changing_default_depot_works(self):
        DepotManager.configure('first', {'depot.storage_path': './lfs'})
        DepotManager.configure('second', {'depot.storage_path': './lfs2'})
        DepotManager.set_default('second')
        assert DepotManager.get_default() == 'second'

    def test_no_configured_is_detected(self):
        with self.assertRaises(RuntimeError):
            DepotManager.get_default()

    def test_prevent_non_existing_default(self):
        with self.assertRaises(RuntimeError):
            DepotManager.set_default('does_not_exists')

    def test_prevent_multiple_middlewares(self):
        with self.assertRaises(RuntimeError):
            DepotManager.make_middleware(None)
            DepotManager.make_middleware(None)

    def test_detects_no_middleware(self):
        with self.assertRaises(RuntimeError):
            DepotManager.get_middleware()

    def test_prevent_configuring_two_storages_with_same_name(self):
        DepotManager.configure('first', {'depot.storage_path': './lfs'})

        with self.assertRaises(RuntimeError):
            DepotManager.configure('first', {'depot.storage_path': './lfs2'})

    def test_aliases(self):
        DepotManager.configure('first', {'depot.storage_path': './lfs'})
        DepotManager.configure('second', {'depot.storage_path': './lfs2'})

        DepotManager.alias('used_storage', 'first')
        storage = DepotManager.get('used_storage')
        assert storage.storage_path == './lfs', storage

        DepotManager.alias('used_storage', 'second')
        storage = DepotManager.get('used_storage')
        assert storage.storage_path == './lfs2', storage

    def test_aliases_not_existing(self):
        with self.assertRaises(ValueError):
            DepotManager.alias('used_storage', 'first')

    def test_alias_on_existing_storage(self):
        DepotManager.configure('mystorage', {'depot.storage_path': './lfs2'})
    
        with self.assertRaises(ValueError):
            DepotManager.alias('mystorage', 'mystorage')




"""
Provides FileStorage implementation for Amazon S3 through boto3 library.

This is useful for storing files in S3.

"""
from __future__ import absolute_import

from datetime import datetime
import uuid
import boto3
from botocore.exceptions import ClientError
from depot._compat import unicode_text, percent_encode, percent_decode
from depot.utils import make_content_disposition

from .interfaces import FileStorage, StoredFile
from . import utils

CANNED_ACL_PUBLIC_READ = 'public-read'
CANNED_ACL_PRIVATE = 'private'


class S3StoredFile(StoredFile):
    def __init__(self, file_id, key):
        _check_file_id(file_id)
        self._closed = False
        self._key = key
        self._body = None
        filename = key.metadata.get('x-depot-filename')
        if filename:
            filename = percent_decode(filename)

        metadata_info = {'filename': filename,
                         'content_type': key.content_type,
                         'content_length': key.content_length,
                         'last_modified': None}

        try:
            last_modified = key.metadata.get('x-depot-modified')
            if last_modified:
                metadata_info['last_modified'] = datetime.strptime(last_modified,
                                                                   '%Y-%m-%d %H:%M:%S')
        except:
            pass

        super(S3StoredFile, self).__init__(file_id=file_id, **metadata_info)

    def read(self, n=-1):
        if self.closed:
            raise ValueError("cannot read from a closed file")

        if self._body is None:
            self._body = self._key.get()['Body']

        if n <= 0:
            n = None
        return self._body.read(n)

    def close(self):
        self._closed = True
        if self._body is not None:
            self._body.close()

    @property
    def closed(self):
        return self._closed

    @property
    def public_url(self):
        expires_in = 31536000 # 1 YEAR
        url = self._key.meta.client.generate_presigned_url(
            'get_object',
            Params={'Bucket': self._key.Bucket().name, 'Key': self._key.key},
            ExpiresIn=expires_in
        )
        # Remove query auth
        url = url.split('?', 1)[0]
        return url


class BucketDriver(object):

    def __init__(self, s3, bucket, prefix):
        self.s3 = s3
        self.bucket = bucket
        self.prefix = prefix

    def get_key(self, key_name):
        k = self.bucket.Object('%s%s' % (self.prefix, key_name))
        try:
            k.reload()
        except ClientError as exc:
            k = None
            if exc.response['Error']['Code'] != '404':
                raise
        return k

    def new_key(self, key_name):
        return self.bucket.Object('%s%s' % (self.prefix, key_name))

    def list_key_names(self):
        keys = self.bucket.objects.filter(Prefix=self.prefix)
        return [k.key[len(self.prefix):] for k in keys]


class S3Storage(FileStorage):
    """:class:`depot.io.interfaces.FileStorage` implementation that stores files on S3.

    **This is a version implemented on top of boto3**.
    Installing ``boto3`` as a dependency is required to use this.

    All the files are stored inside a bucket named ``bucket`` on ``host`` which Depot
    connects to using ``access_key_id`` and ``secret_access_key``.
 
    Additional options include:
        * ``region`` which can be used to specify the AWS region.
        * ``endpoint_url`` which can be used to specify an host different from Amazon
          AWS S3 Storage
        * ``policy`` which can be used to specify a canned ACL policy of either
          ``private`` or ``public-read``.
        * ``storage_class`` which can be used to specify a class of storage.
        * ``prefix`` parameter can be used to store all files under 
          specified prefix. Use a prefix like **dirname/** (*see trailing slash*)
          to store in a subdirectory.
    """

    def __init__(self, access_key_id, secret_access_key, bucket=None, region_name=None,
                 policy=None, storage_class=None, endpoint_url=None, prefix=''):
        policy = policy or CANNED_ACL_PUBLIC_READ
        assert policy in [CANNED_ACL_PUBLIC_READ, CANNED_ACL_PRIVATE], (
            "Key policy must be %s or %s" % (CANNED_ACL_PUBLIC_READ, CANNED_ACL_PRIVATE))
        self._policy = policy or CANNED_ACL_PUBLIC_READ
        self._storage_class = storage_class or 'STANDARD'

        if bucket is None:
            bucket = 'filedepot-%s' % (access_key_id.lower(),)

        kw = {}
        if endpoint_url is not None:
            kw['endpoint_url'] = endpoint_url
        if region_name is not None:
            kw['region_name'] = region_name
        self._conn = boto3.Session(aws_access_key_id=access_key_id,
                                   aws_secret_access_key=secret_access_key)
        self._s3 = self._conn.resource('s3', **kw)
        bucket = self._s3.Bucket(bucket)

        # Create bucket if it doesn't exist.
        buckets = set(b['Name'] for b in self._s3.meta.client.list_buckets()['Buckets'])
        if bucket.name not in buckets:
            bucket.create(ObjectOwnership="BucketOwnerPreferred")
            bucket.wait_until_exists()
            if self._policy == CANNED_ACL_PUBLIC_READ:
                self._conn.client('s3').put_public_access_block(
                    Bucket=bucket.name,
                    PublicAccessBlockConfiguration={
                        'BlockPublicAcls': False,
                        'IgnorePublicAcls': False,
                        'BlockPublicPolicy': True,
                        'RestrictPublicBuckets': True
                    }
                )

        self._bucket_driver = BucketDriver(self._s3, bucket, prefix)

    def get(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        key = self._bucket_driver.get_key(fileid)
        if key is None:
            raise IOError('File %s not existing' % fileid)

        return S3StoredFile(fileid, key)

    def __save_file(self, key, content, filename, content_type=None):
        if filename:
            filename = percent_encode(filename, safe='!#$&+-.^_`|~', encoding='utf-8')
        attrs = {
            'ACL': self._policy,
            'StorageClass': self._storage_class,
            'Metadata': {
                'x-depot-filename': filename,
                'x-depot-modified': utils.timestamp()
            },
            'ContentType': content_type,
            'ContentDisposition': make_content_disposition('inline', filename)
        }

        if hasattr(content, 'read'):
            try:
                pos = content.tell()
                content.seek(pos)
            except:
                content = content.read()
            key.put(Body=content, **attrs)
        else:
            if isinstance(content, unicode_text):
                raise TypeError('Only bytes can be stored, not unicode')
            key.put(Body=content, **attrs)

    def create(self, content, filename=None, content_type=None):
        content, filename, content_type = self.fileinfo(content, filename, content_type)
        new_file_id = str(uuid.uuid1())
        key = self._bucket_driver.new_key(new_file_id)
        self.__save_file(key, content, filename, content_type)
        return new_file_id

    def replace(self, file_or_id, content, filename=None, content_type=None):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        if isinstance(file_or_id, StoredFile) and file_or_id is content:
            # This is a backup, no need to check if file exists.
            pass
        elif not self.exists(fileid):
            # Check file existed and we are not using replace
            # as a way to force a specific file id on creation.
            raise IOError('File %s not existing' % fileid)

        content, filename, content_type = self.fileinfo(content, filename, content_type,
                                                        lambda: self.get(fileid))

        key = self._bucket_driver.get_key(fileid)
        if key is None:
            key = self._bucket_driver.new_key(fileid)

        self.__save_file(key, content, filename, content_type)
        return fileid

    def delete(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        k = self._bucket_driver.get_key(fileid)
        if k:
            k.delete()

    def exists(self, file_or_id):
        fileid = self.fileid(file_or_id)
        _check_file_id(fileid)

        k = self._bucket_driver.get_key(fileid)
        return k is not None

    def list(self):
        return self._bucket_driver.list_key_names()


def _check_file_id(file_id):
    # Check that the given file id is valid, this also
    # prevents unsafe paths.
    try:
        uuid.UUID('{%s}' % file_id)
    except:
        raise ValueError('Invalid file id %s' % file_id)



from ._compat import percent_encode, unicode_text

try:
    from anyascii import anyascii
except ImportError:  # pragma: no cover
    # Python2 doesn't support anyascii
    import unicodedata
    def anyascii(text):
        if not isinstance(text, unicode_text):
            text = text.decode("utf-8")
        return unicodedata.normalize("NFKD", text).encode("ascii", "ignore") or "unknown"


def make_content_disposition(disposition, fname):
    rfc6266_part = "filename*=utf-8''%s" % (percent_encode(fname, safe='!#$&+-.^_`|~', encoding='utf-8'), )
    ascii_part = 'filename="%s"' % (anyascii(fname), )
    return ';'.join((disposition, ascii_part, rfc6266_part))
